/*
 * (C) 2002 Clemson University and The University of Chicago.
 *
 * See COPYING in top-level directory.
 */

#ifndef __CINT_DISTRIBUTION_H
#define __CINT_DISTRIBUTION_H

/* modes for CINT_Process_request */
#define CINT_SERVER 1
#define CINT_CLIENT 2
#define CINT_CKSIZE 3

/* COVEN Type Processing Stuff */

typedef struct CINT_Type {
	COVEN_offset  offset;        /* offset from start of last set of elements */
	COVEN_count32 num_etypes;     /* number of ereqs in a block */
	COVEN_size    stride;        /* stride between blocks in bytes */
	COVEN_count32 num_blocks;    /* number of blocks */
	COVEN_offset  ub;            /* upper bound of the type in bytes */
	COVEN_offset  lb;            /* lower bound of the type in bytes */
	COVEN_size    aggregate_size; /* amount of aggregate data in bytes */
	COVEN_count32 depth;    	    /* number of levels of nesting */
	COVEN_count32 num_contig_chunks; /* number of contiguous data chunks */
	struct CINT_Type *etype;  /* element type */
	struct CINT_Type *stype;  /* sequence type */
} CINT_Type;

typedef struct CINT_typestack {
	COVEN_count32 el;           /* number of element being processed */
	COVEN_count32 maxel;        /* total number of these elements to process */
	CINT_Type *ty;    		/* pointer to request structure */
	CINT_Type *tybase; 		/* pointer to first request is sequence chain */
	COVEN_count32 blk;          /* number of block being processed */
	COVEN_offset  chunk_offset; /* offset of beginning of current contiguous chunk */
} CINT_typestack;           
          
typedef struct CINT_Type_state { 
	struct CINT_typestack *cur; /* request element chain stack */
	COVEN_count32 lvl;          /* level in element chain */
	COVEN_size    bytes;        /* bytes in current contiguous chunk processed */
	COVEN_offset  buf_offset;   /* byte offset in user buffer */
} CINT_Type_state;           

typedef struct CINT_Type_file_data {
	COVEN_size    fsize;			/* actual size of local storage object */
	COVEN_count32 iod_num;		/* ordinal number of THIS server for this file */
	COVEN_count32 iod_count;		/* number of servers for this file */
	// COVEN_Distribution *dist;
	// COVEN_Dist_parm    *dparm;
	COVEN_boolean      extend_flag;
} CINT_Type_file_data;

struct CINT_Type_state *CINT_New_datatype_state (CINT_Type *datatype);

void CINT_Free_datatype_state (CINT_Type_state *type_state);

int CINT_Process_type(CINT_Type_state *type_state,
	   COVEN_count tph_count,
	   COVEN_TPH *tph,
		COVEN_Buffer buffer,
	   COVEN_flag mode);

COVEN_size CINT_Distribute(COVEN_offset offset, COVEN_size size,
      COVEN_size *bytes,
      COVEN_count32 *segs,
      void *buffer,
      void *buffer2,
      COVEN_flag mode);

int CINT_Scatter(CINT_Type *datatype, COVEN_count tph_count,
  COVEN_TPH *tph, COVEN_Buffer buffer);
int CINT_Gather(CINT_Type *datatype, COVEN_count tph_count,
  COVEN_TPH *tph, COVEN_Buffer buffer);
void* CINT_create_buffer(struct COVEN_TPH *tph,
  int num_bufs, int this_buf_num, char *tag, int ptype, int psize);
void* CINT_get_buffer(struct COVEN_TPH *tph,
  int *num_bufs, int *this_buf_num, char *tag, int ptype);

#endif /* __CINT_DISTRIBUTION_H */
