#ifndef __COVEN_H
#define __COVEN_H

#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include "mpi.h"
#include "coven_debug.h"
#include "sockio.h"
#include <sys/time.h>
#include <unistd.h>
//#include "PSE_types.h"

#include <gossip.h>
#include "coven_structs.h"
#include "coven_prototypes.h"
#include "coven_datatype.h"

/* BEGIN: MPI message tags */
#define COVEN_PROGRAM_SEQUENCER_TAG	10
/* END: MPI message tags */

/* BEGIN: COVEN specific messages */
#define COVEN_SHUTDOWN_MESSAGE	110
#define COVEN_TPH_QUEUED_MESSAGE 111
/* END: COVEN specific messages */

#define PSE_TYPE_OFFSET 100

MPI_Comm SLAVE_COMM;

#define TRUE 1
#define FALSE 0

#define MIN(x,y) (x < y) ? x : y

#define SLAVE 1
#define MASTER 0

/* COVEN magic number used for memory operations */
#define COVEN_MAGIC	666

/* COVEN debug defines for gossip */
#define DATATYPE_DEBUG	1
#define DLOOP 		2
#define DMAIN		4
#define DRESULTS	8
#define DCTRL		16
#define DQUEUE		32
#define DMODULE		64
#define DTPH		128
#define DARGS		256
#define DBND		512
#define DMEM		1024
#define DMEMV		2048
#define DMODS		4096
#define DMARSHALL	8192
#define DCOL		16384
#define DSPECIAL	32768
/* COVEN looping stuff */
#define COVEN_LT 12
#define COVEN_LEQ 13
#define COVEN_GT 14
#define COVEN_GEQ 15
#define COVEN_EQ 16
#define COVEN_NEQ 17

/* COVEN Attribute Types */
#define COVEN_CHAR  		0
#define COVEN_UCHAR 		1  
#define COVEN_SHORT 		2
#define COVEN_USHORT		3
#define COVEN_INT   		4
#define COVEN_UINT  		5
#define COVEN_FLOAT 		6
#define COVEN_DOUBLE		7
#define COVEN_STRING		8

/* COVEN Buffer Modes */
#define COVEN_INPUT_MODE		0
#define COVEN_OUTPUT_MODE		1
#define COVEN_INOUT_MODE		2

/* COVEN Attribute Modes */
#define COVEN_ATTRIB_GET_MODE	10
#define COVEN_ATTRIB_PUT_MODE	11

#define COVEN_Buffer_Replace(local_binding, ptr, num) \
	COVEN_Do_Buffer_Data_Replace(local_binding, _default_tph, btable, \
	  ptr, num)

#define COVEN_Create_Scatter_Action(act, dt, local_binding) \
	COVEN_Do_Create_Action(act, dt, local_binding, _default_tph, \
	  btable, Scatter)

#define COVEN_Create_Scatter_Action_Static(act, dt, local_binding) \
	COVEN_Do_Create_Action(act, dt, local_binding, _static_tph, \
	  btable, Scatter)

#define COVEN_Scatter(num, act, num_actions) \
	COVEN_Do_Scatter(_default_tph, num, act, num_actions)

#define COVEN_Create_Gather_Action(act, dt, local_binding) \
   	COVEN_Do_Create_Action(act, dt, local_binding, _default_tph, \
	  btable, Gather)
   
#define COVEN_Gather(num, act, num_actions) \
	COVEN_Do_Gather(_default_tph, _static_tph, num, act, num_actions)

#define COVEN_MODULE(fname, deftph, consttph, statictph) \
	int fname (struct COVEN_TPH *deftph, struct COVEN_TPH *consttph, \
	  struct COVEN_TPH *statictph, struct COVEN_Mod_Binding_Table *btable)

#define COVEN_INPUT(tphname, itype, local_binding) \
	COVEN_Get_Buffer_Arg(tphname, btable, local_binding, COVEN_INPUT_MODE, itype, 0)

#define COVEN_OUTPUT(tphname, otype, osize, local_binding) \
	COVEN_Get_Buffer_Arg(tphname, btable, local_binding, COVEN_OUTPUT_MODE, otype, osize)

#define COVEN_INOUT(tphname, iotype, local_binding) \
	COVEN_Get_Buffer_Arg(tphname, btable, local_binding, COVEN_INOUT_MODE, iotype, 0)

/* BEGIN: Attribute get and put macros */
#define ATTRIB_GET(at, local_binding) \
	COVEN_Get_Attrib_Arg(at, btable, local_binding)

#define ATTRIB_PUT(at, local_binding, type, size) \
   	COVEN_Put_Attrib_Arg(at, btable, local_binding, type, size)

#define ATTRIB_GET_PTR(at, local_binding, type, size) \
   	COVEN_Put_Ptr_Attrib_Arg(at, btable, local_binding, type, size)

#define COVEN_CHAR_CONST(local_binding) \
	*(char*)(ATTRIB_GET(_const_tph->at, local_binding))
#define COVEN_CHAR_ATTRIB_GET(tphname, local_binding) \
	*(char*)(ATTRIB_GET(tphname->at, local_binding))
#define COVEN_CHAR_ATTRIB_GET_PTR(tphname, local_binding) \
	(char*)(ATTRIB_GET_PTR(tphname->at, local_binding, COVEN_CHAR, 0))
#define COVEN_CHAR_ATTRIB_PUT(tphname, local_binding) \
	(char*)(ATTRIB_PUT(tphname->at, local_binding, COVEN_CHAR, 0))

#define COVEN_UCHAR_CONST(local_binding) \
	*(unsigned char*)(ATTRIB_GET(_const_tph->at, local_binding))
#define COVEN_UCHAR_ATTRIB_GET(tphname, local_binding) \
	*(unsigned char*)(ATTRIB_GET(tphname->at, local_binding))
#define COVEN_UCHAR_ATTRIB_GET_PTR(tphname, local_binding) \
	(unsigned char*)(ATTRIB_GET_PTR(tphname->at, local_binding, \
	  COVEN_UCHAR, 0))
#define COVEN_UCHAR_ATTRIB_PUT(tphname, local_binding) \
	(unsigned char*)(ATTRIB_PUT(tphname->at, local_binding, COVEN_UCHAR, 0))

#define COVEN_SHORT_CONST(local_binding) \
	*(short*)(ATTRIB_GET(_const_tph->at, local_binding))
#define COVEN_SHORT_ATTRIB_GET(tphname, local_binding) \
	*(short*)(ATTRIB_GET(tphname->at, local_binding))
#define COVEN_SHORT_ATTRIB_GET_PTR(tphname, local_binding) \
	(short*)(ATTRIB_GET_PTR(tphname->at, local_binding, COVEN_SHORT, 0))
#define COVEN_SHORT_ATTRIB_PUT(tphname, local_binding) \
	(short*)(ATTRIB_PUT(tphname->at, local_binding, COVEN_SHORT, 0))

#define COVEN_USHORT_CONST(local_binding) \
	*(unsigned short*)(ATTRIB_GET(_const_tph->at, local_binding))
#define COVEN_USHORT_ATTRIB_GET(tphname, local_binding) \
	*(unsigned short*)(ATTRIB_GET(tphname->at, local_binding))
#define COVEN_USHORT_ATTRIB_GET_PTR(tphname, local_binding) \
	(unsigned short*)(ATTRIB_GET_PTR(tphname->at, local_binding, \
	  COVEN_USHORT, 0))
#define COVEN_USHORT_ATTRIB_PUT(tphname, local_binding) \
	(unsigned short*)(ATTRIB_PUT(tphname->at, local_binding, COVEN_USHORT, 0))

#define COVEN_INT_CONST(local_binding) \
	*(int*)(ATTRIB_GET(_const_tph->at, local_binding))
#define COVEN_INT_ATTRIB_GET(tphname, local_binding) \
	*(int*)(ATTRIB_GET(tphname->at, local_binding))
#define COVEN_INT_ATTRIB_GET_PTR(tphname, local_binding) \
	(int*)(ATTRIB_GET_PTR(tphname->at, local_binding, COVEN_INT, 0))
#define COVEN_INT_ATTRIB_PUT(tphname, local_binding) \
	(int*)(ATTRIB_PUT(tphname->at, local_binding, COVEN_INT, 0))

#define COVEN_UINT_CONST(local_binding) \
	*(unsigned int*)(ATTRIB_GET(_const_tph->at, local_binding))
#define COVEN_UINT_ATTRIB_GET(tphname, local_binding) \
	*(unsigned int*)(ATTRIB_GET(tphname->at, local_binding))
#define COVEN_UINT_ATTRIB_GET_PTR(tphname, local_binding) \
	(unsigned int*)(ATTRIB_GET_PTR(tphname->at, local_binding, \
	  COVEN_UINT, 0))
#define COVEN_UINT_ATTRIB_PUT(tphname, local_binding) \
	(unsigned int*)(ATTRIB_PUT(tphname->at, local_binding, COVEN_UINT, 0))

#define COVEN_FLOAT_CONST(local_binding) \
	*(float*)(ATTRIB_GET(_const_tph->at, local_binding))
#define COVEN_FLOAT_ATTRIB_GET(tphname, local_binding) \
	*(float*)(ATTRIB_GET(tphname->at, local_binding))
#define COVEN_FLOAT_ATTRIB_GET_PTR(tphname, local_binding) \
	(float*)(ATTRIB_GET_PTR(tphname->at, local_binding, COVEN_FLOAT, 0))
#define COVEN_FLOAT_ATTRIB_PUT(tphname, local_binding) \
	(float*)(ATTRIB_PUT(tphname->at, local_binding, COVEN_FLOAT, 0))

#define COVEN_DOUBLE_CONST(local_binding) \
	*(double*)(ATTRIB_GET(_const_tph->at, local_binding))
#define COVEN_DOUBLE_ATTRIB_GET(tphname, local_binding) \
	*(double*)(ATTRIB_GET(tphname->at, local_binding))
#define COVEN_DOUBLE_ATTRIB_GET_PTR(tphname, local_binding) \
	(double*)(ATTRIB_GET_PTR(tphname->at, local_binding, COVEN_DOUBLE, 0))
#define COVEN_DOUBLE_ATTRIB_PUT(tphname, local_binding) \
	(double*)(ATTRIB_PUT(tphname->at, local_binding, COVEN_DOUBLE, 0))

#define COVEN_STRING_CONST(local_binding) \
	(unsigned char*)(ATTRIB_GET(_const_tph->at, local_binding))
#define COVEN_STRING_ATTRIB_GET(tphname, local_binding) \
	(unsigned char*)(ATTRIB_GET(tphname->at, local_binding))
/* duh, a string PTR or regular is still a ptr */
#define COVEN_STRING_ATTRIB_GET_PTR(tphname, local_binding) \
	(unsigned char*)(ATTRIB_GET_PTR(tphname->at, local_binding, \
	  COVEN_STRING, length))
/* notice that STRING ATTRIB PUT has a length (size) of the string */
#define COVEN_STRING_ATTRIB_PUT(tphname, local_binding, length) \
	(unsigned char*)(ATTRIB_PUT(tphname->at, local_binding, COVEN_STRING, length))
/* END: attribute get and put macros */

#define COVEN_NEW_TPH() \
	COVEN_New_TPH()
	
#define COVEN_FORK_TPH(tphname) \
	COVEN_Copy_TPH(tphname)

#define COVEN_PUSH_OLD_CLASS_INFO(tphname) \
	COVEN_Push_Old_Class_Info(tphname)

#define COVEN_POP_OLD_CLASS_INFO(tphname) \
	COVEN_Pop_Old_Class_Info(tphname)

#define COVEN_COPY_ATTRIBUTE_TABLE(oldtphname, newtphname) \
	COVEN_Copy_Attribute_Table(oldtphname, newtphname)

#define COVEN_RELAY_TPH(tphname) \
	COVEN_Relay_TPH_Simple(tphname)

#define COVEN_SCHEDULE_TPH_FOR_DEALLOC(tphname) \
	COVEN_Schedule_TPH_For_Dealloc(tphname)

#define COVEN_MOVE_BUFFER(bufdata, oldtph, newtph) \
	COVEN_Move_Buffer((void**)&bufdata, oldtph, newtph)

#define COVEN_SIZE(pname) \
	COVEN_Size(pname)

#define COVEN_COUNT(pname) \
	COVEN_Count(pname)

#define COVEN_ERROR(string) \
	COVEN_Error(string)

#define COVEN_NEW_BUFFER(tph, tag, size, type) \
	COVEN_New_Buffer(tph, tag, size, type)

#define COVEN_GET_BUFFER_DATA(tph, tag, type) \
	COVEN_Get_Buffer_Data(tph, tag, type)

#endif /* __COVEN_H */
