#ifndef __COVEN_PROTOTYPES_H
#define __COVEN_PROTOTYPES_H
#include <coven_structs.h>

/* prototypes */

/* COVEN_Profile.c */
void COVEN_Profile_Enable(void);
void COVEN_Profile_Disable(void);
COVEN_Profile_Stamp * COVEN_Profile_Stamp_Malloc(void);
void COVEN_Profile_Stamp_Free(COVEN_Profile_Stamp *ps);
COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Module(struct COVEN_TPH *tph,
	struct COVEN_Mod_Call *mod_call);
COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Module_In(struct COVEN_TPH *tph,
	struct COVEN_Mod_Call *mod_call);
COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Module_Out(struct COVEN_TPH *tph,
	struct COVEN_Mod_Call *mod_call);
COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Queue(struct COVEN_TPH *tph,
	struct COVEN_TPH_Queue *q);
COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Queue_In(struct COVEN_TPH *tph,
	struct COVEN_TPH_Queue *q);
COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Queue_Out(struct COVEN_TPH *tph,
	struct COVEN_TPH_Queue *q);
COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Created(struct COVEN_TPH *tph);
COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Destroyed(struct COVEN_TPH *tph);
COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Sleep(struct COVEN_TPH *tph);
COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Awaken(struct COVEN_TPH *tph);
void COVEN_Profile_Stamp_TPH(COVEN_Profile_Stamp *ps, struct COVEN_TPH *tph);
	
void COVEN_Profile_Stamp_Copy(COVEN_Profile_Stamp *src, COVEN_Profile_Stamp *dst);
void COVEN_Profile_Table_Append_Stamp(COVEN_Profile_Table *pt, COVEN_Profile_Stamp *new_ps);
void COVEN_Profile_Table_Print(COVEN_Profile_Table *pt, FILE *stream);
void COVEN_Profile_Table_Write(struct COVEN_TPH *tph);
void COVEN_Profile_Table_Free(COVEN_Profile_Table *pt);
char * COVEN_Profile_Stamp_Type_String(_stamp_type st);

/* COVEN_IO.c */
int COVEN_write(int fd, const void *buf, int count);
int COVEN_read(int fd, void *buf, int count);

/* COVEN_attributes.c */
int COVEN_Find_Attribute_Next_Number(struct COVEN_TPH *tph, struct COVEN_Attribute *ptr);
int COVEN_Modify_Attribute_Name(struct COVEN_TPH *tph, struct COVEN_Attribute *ptr);
void COVEN_Copy_Attribute_Table(struct COVEN_TPH *oldtph, struct COVEN_TPH *newtph);
void COVEN_Change_Attrib_Tag(struct COVEN_Attribute_Table *at, char *oldtag, char *newtag);
struct COVEN_Attribute *COVEN_New_Attribute(int type, char *tag, void *data);
void COVEN_Print_Attribute(struct COVEN_Attribute *ptr);
void COVEN_Print_Attribute_Table(struct COVEN_Attribute_Table *at);
void *COVEN_Attribute_Table_Put_Attrib_Data(struct COVEN_Attribute_Table *at, char *tag, int type, int size);
void *COVEN_Attribute_Table_Get_Attrib_Data(struct COVEN_Attribute_Table *at, char *tag);
void *COVEN_Attribute_Table_Get_Or_Create_Attrib_Data(
  struct COVEN_Attribute_Table *at, char *tag, int type, int size);
int COVEN_Attribute_Table_Append_Attrib(struct COVEN_Attribute_Table *at, struct COVEN_Attribute *new_at);
int COVEN_Attribute_Table_Force_Attrib(struct COVEN_Attribute_Table *at, struct COVEN_Attribute *new_at);
int COVEN_Attribute_Table_Delete_Attrib(struct COVEN_Attribute_Table *at, struct COVEN_Attribute *find_me);
struct COVEN_Attribute *COVEN_Attribute_Table_Find_Attrib(struct COVEN_Attribute_Table *at, char *find_tag);
void COVEN_Free_Attribute_Table(struct COVEN_Attribute_Table *at);

/* COVEN_buffers.c */
void * COVEN_Buffer_Data(struct COVEN_Buffer *buf);
void COVEN_Put_Buffer(struct COVEN_TPH *tph, struct COVEN_Buffer *buf);
void COVEN_Move_Buffer(void **bufdata, struct COVEN_TPH *oldtph, struct COVEN_TPH *newtph);
struct COVEN_Buffer * COVEN_Do_Buffer_Data_Replace(char *local_binding, 
  struct COVEN_TPH *tph, struct COVEN_Mod_Binding_Table *btable, 
  void *new_data, int num_elements);
void COVEN_Do_Buffer_Replace(struct COVEN_TPH *tph,
  struct COVEN_Buffer *orig_buf, struct COVEN_Buffer *new_buf);
void COVEN_Free_Buffer(struct COVEN_Buffer *buffer);
struct COVEN_Buffer * COVEN_Get_Buffer(struct COVEN_TPH *tph, char *find_tag, int ptype);
struct COVEN_Buffer * COVEN_New_Buffer(struct COVEN_TPH *tph, char *tag, int psize, int ptype);
struct COVEN_Buffer * COVEN_Just_Create_Buffer(struct COVEN_TPH *tph, char *tag, int psize, int ptype);
void * COVEN_Get_Buffer_Data(struct COVEN_TPH *tph, char *tag, int ptype);
int COVEN_Size(void *data);
int COVEN_Count(void *data);

/* COVEN_dynamic_loading.c */
void COVEN_Fill_TPH_With_Symbols(struct COVEN_TPH *tph);
void * COVEN_Open_Dynamics(void);
void COVEN_Close_Dynamics(void *dl);
void COVEN_Load_Dynamics(void *dl, struct COVEN_Tag_Usage *tagusage);
void * COVEN_dl_get(void *dl, char *symbol);
struct COVEN_TPH * COVEN_Setup_Constant_TPH(struct COVEN_Consts *consts);

/* COVEN_marshalling.c */
void COVEN_Marshall_TPH(struct COVEN_TPH *tph, void **marshalled_TPH, int *marshalled_TPH_size);
struct COVEN_TPH * COVEN_Unmarshall_TPH(void *marshalled_TPH);

/* COVEN_misc.c */
void COVEN_Error(char *str);
void COVEN_Error(char *str);

/* COVEN_module_args.c */
void * COVEN_Get_Buffer_Arg(struct COVEN_TPH *tph, 
  struct COVEN_Mod_Binding_Table *btable, char *local_binding, 
  int pmode, int ptype, int psize);
void * COVEN_Put_Attrib_Arg(struct COVEN_Attribute_Table *at,
  struct COVEN_Mod_Binding_Table *btable, char *local_binding, int type,
  int size);
void * COVEN_Put_Ptr_Attrib_Arg(struct COVEN_Attribute_Table *at,
  struct COVEN_Mod_Binding_Table *btable, char *local_binding, int type,
  int size);
void * COVEN_Get_Attrib_Arg(struct COVEN_Attribute_Table *at, 
  struct COVEN_Mod_Binding_Table *btable, char *local_binding);

/* COVEN_program_sequencer.c */
void COVEN_Master_Inject(void);
void COVEN_Shut_Down(void);
void COVEN_Program_Sequencer(void);
void COVEN_process_received(struct COVEN_TPH *tph);
void COVEN_Get_Next_Task(struct COVEN_TPH *tph, int *proc_type, int *task_id);
int COVEN_Run_Module(struct COVEN_TPH *tph, struct COVEN_TPH *ctph, 
  struct COVEN_TPH *stph, struct COVEN_Mod_Call *mcall);
void COVEN_Print_Module(struct COVEN_Mod_Call *mod);
void COVEN_Malloc_Static_TPHs(void);
int COVEN_Process_Loop(struct COVEN_Loop *loop, struct COVEN_TPH *tph,
  struct COVEN_TPH *ctph, int this_mod_number);

/* COVEN_tag_usage.c */
void COVEN_Update_Tag_Usage(char *tag, int stepnum, struct COVEN_Tag_Usage *tagusage);
int COVEN_Find_Last_Used(char *tag, struct COVEN_Tag_Usage *tagusage);
struct COVEN_Tag_Usage* COVEN_Find_Tag_Usage(char *tag, struct COVEN_Tag_Usage *tagusage);
void COVEN_Add_Tag_Usage(char *tag, int stepnum, struct COVEN_Tag_Usage *tagusage);
void COVEN_Print_Tag_Usage(struct COVEN_Tag_Usage *tagusage);

/* COVEN_TPH.c */
void COVEN_Print_TPH(struct COVEN_TPH *tph);
void COVEN_Setup_TPH(struct COVEN_TPH *tph);
struct COVEN_TPH * COVEN_New_TPH(void);
struct COVEN_TPH * COVEN_Copy_TPH(struct COVEN_TPH *tph);
struct COVEN_TPH * COVEN_Copy_TPH_Exclude(struct COVEN_TPH *tph,
  struct COVEN_Collective_Action *actions, int num_actions);
void COVEN_Free_TPH(struct COVEN_TPH *tph);
void COVEN_Cleanup_TPH(struct COVEN_TPH *tph, int completed_stepnum, struct COVEN_Tag_Usage *tagusage);
void COVEN_Schedule_TPH_For_Dealloc(struct COVEN_TPH *tph);
int COVEN_TPH_Scheduled_For_Dealloc(struct COVEN_TPH *tph);
struct COVEN_TPH * COVEN_Get_TPH(void *data);
void COVEN_TPH_Clear_Manual_Relay(struct COVEN_TPH *tph);
int COVEN_TPH_Manually_Relayed(struct COVEN_TPH *tph);
void COVEN_Manually_Relay(struct COVEN_TPH *tph, int thread_number);
void COVEN_Relay_TPH_Locally(struct COVEN_TPH *tph, int proc_type, int task_id);
void COVEN_Relay_TPH(struct COVEN_TPH *tph, int proc_type, int proc_num, int task_id);
void COVEN_TPH_Get_Stats(struct COVEN_TPH *tph, int *num_tags, int *totals_bytes);
int COVEN_Convert_To_MPI_TID(int proc_type, int proc_num, int task_id);
void COVEN_Relay_TPH_Simple(struct COVEN_TPH *tph);
int COVEN_Scheduler_Get_Next(void);

/* COVEN_TPH_queues.c */
void COVEN_Enqueue_TPH(struct COVEN_TPH *tph);
struct COVEN_TPH * COVEN_Dequeue_TPH(void);

/* COVEN_types.c */
int COVEN_get_type_size(int which);
char * COVEN_get_mode_string(int which);

/* COVEN_main.c */
void master(char *working_dir);
void slave(void);

/* COVEN_binding.c */
char * COVEN_Get_Global_Binding(struct COVEN_Mod_Binding_Table *btable, 
  char *local_binding);

/* COVEN_helpers.c */
void * COVEN_malloc(int num_bytes);
void COVEN_free(void *ptr);
char * COVEN_Get_Debug_Prepend();
void COVEN_Print_Program(struct COVEN_Program *pgm);
void COVEN_Print_Consts(struct COVEN_Consts *consts);
void COVEN_Print_Const(struct COVEN_Const_Decl *cd);
void COVEN_Print_Mods(struct COVEN_Mods *mods);
void COVEN_Print_Mod_Decl(struct COVEN_Mod_Decl *md);
void COVEN_Print_Interface(struct COVEN_Interface *iface);
void COVEN_Print_Interface_Element(struct COVEN_Interface_Element *iel);
void COVEN_Print_Mod_Calls(struct COVEN_Mod_Calls *calls);
void COVEN_Print_Mod_Call(struct COVEN_Mod_Call *call);
void COVEN_Print_Binding_Table(struct COVEN_Mod_Binding_Table *btable);
void COVEN_Print_Binding(struct COVEN_Mod_Binding *binding);

/* coven_collective.c */
void COVEN_Do_Create_Action(struct COVEN_Collective_Action *action,
  COVEN_Type datatype, char *local_binding, struct COVEN_TPH *tph,
  struct COVEN_Mod_Binding_Table *btable, COVEN_Collective_Type ctype);
void COVEN_Do_Scatter(struct COVEN_TPH *copy_tph, int num_tphs, 
  struct COVEN_Collective_Action *actions, int num_actions);
void COVEN_Do_Gather(struct COVEN_TPH *from_tph, struct COVEN_TPH *to_tph,
  int num_elements, struct COVEN_Collective_Action *actions, int num_actions);

/* PSE/src/PSE_types.c */
int PSE_get_type_size(int which);

#endif /* __COVEN_PROTOTYPES_H */
