#ifndef __COVEN_STRUCTS_H
#define __COVEN_STRUCTS_H

#include <sys/time.h>
#include <unistd.h>
#include "coven_datatype.h"
#include "cint_datatype.h"

struct COVEN_Mod_Decl {
	char *file_name;
	char *module_name;
	struct COVEN_Interface *interface;
};

struct COVEN_Mods {
	int num_mod_decls;
	struct COVEN_Mod_Decl *mod_decls;
};

struct COVEN_Const_Decl {
	int type;
	char *global_binding;
	char *str_value;
};

struct COVEN_Consts {
	int num_consts;
	struct COVEN_Const_Decl *const_decls;
};

struct COVEN_Mod_Binding {
	char *local_binding;
	char *global_binding;
};

struct COVEN_Mod_Binding_Table {
	int num_bindings;
	int try_here;
	struct COVEN_Mod_Binding *bindings;
};

struct COVEN_Interface_Element {
	char *local_binding;
	int type;
	int mode;
};

struct COVEN_Interface {
	int (*func)();
	int num_iface_elems;
	struct COVEN_Interface_Element *iface_elems;
};

/* this struct is used to describe a loop.  We can infer whether, when this
 * loop is executed (at the top of bottom of it) if the executer is the top
 * or bottom because that module number will match either the begin_module or
 * the end_module */
struct COVEN_Loop {
	char * loop_tag; /* tag to place use for the for loop */
	int initial_val; /* the initial value of the tag */
	int cmp_type; /* type of comparison - leq, lt, gt, ne, etc */
	char * cmp_tag; /* tag to test if 'tag' is less than or eq to */
	int incr_val; /* value to increment tag by each time */
	int begin_module; /* module number for the beginning of the loop */
	int end_module; /* module number for the end of the loop */
};

struct COVEN_Mod_Call {
	char *module_name;
	int slave_module; /* is this a slave module or master? */
	int task_number; /* which task does this run under? */
	struct COVEN_Mod_Binding_Table *btable;
	struct COVEN_Loop *loop; /* whether or not this mod has a loop */
	struct COVEN_Interface *interface;
	/* do we need this anymore? */
	int program_number; 
};

struct COVEN_Mod_Calls {
	int num_mod_calls;
	struct COVEN_Mod_Call *mod_calls;
};

struct COVEN_Program {
	int num_master_tasks;
	int num_slave_tasks;
	struct COVEN_Consts *constants;
	struct COVEN_Mods *modules;
	struct COVEN_Mod_Calls *mod_calls;
};

struct COVEN_Tag_Usage {
	char *tagname;
	int last_used;
	struct COVEN_Tag_Usage *next;
};

struct COVEN_Buffer {
	char *tag;
	struct COVEN_TPH *tph;
	int type;
	int size;
	/* buffer segments are used in scatter / gather.  this requires
	 * knowing which segment number this buffer is and how many segments
	 * there are total: [1->N] */
	int seg_num;
	int num_segs;
	unsigned int data;
};

struct COVEN_Buffer_List_El {
	struct COVEN_Buffer *buffer;
	struct COVEN_Buffer_List_El *next;
};

struct COVEN_Attribute {
	char *tag;
	int attribute_type;
	void *data;
	struct COVEN_Attribute *next;
};

struct COVEN_Attribute_Table {
	struct COVEN_Attribute *head;
	struct COVEN_Attribute *tail;
};

/* BEGIN: Profile stuff */
/* the types of states that a TPH can be stamped */
typedef enum {
	Created,
	Destroyed,
	Queue_In,
	Queue_Out,
	Module_In,
	Module_Out,
	Sleep,
	Awaken
} _stamp_type;

typedef struct _COVEN_Profile_Stamp {
	unsigned int processor_number;
	unsigned int mpirank;
	unsigned int coven_task_id;
	int TPH_num_tags;
	int TPH_num_total_bytes;
	_stamp_type stamp_type;
	unsigned int special_int;
	unsigned char special_string[128];
	struct timeval timestamp;
	struct _COVEN_Profile_Stamp *next;
} COVEN_Profile_Stamp;

typedef struct {
	COVEN_Profile_Stamp *head;
	COVEN_Profile_Stamp *tail;
} COVEN_Profile_Table;
/* END: Profile stuff */

struct COVEN_TPH {
	struct COVEN_Attribute_Table *at;
	COVEN_Profile_Table *pt;
	struct COVEN_Buffer_List_El *head;
	struct COVEN_Buffer_List_El *tail;
};

struct COVEN_TPH_Queue {
	int size;
	struct COVEN_TPH_Queue_Element *back;
	struct COVEN_TPH_Queue_Element *front;
};

struct COVEN_TPH_Queue_Element {
	struct COVEN_TPH *tph;
	struct COVEN_TPH_Queue_Element *infront;
	struct COVEN_TPH_Queue_Element *behind;
};

typedef enum {
	Scatter,
	Gather,
} COVEN_Collective_Type;

struct COVEN_Collective_Action {
	COVEN_Type datatype;
	struct COVEN_Buffer *buffer;
	COVEN_Collective_Type type;
};

/* I think these are a bunch of old structs that we don't need / use anymore.
 * Keep them here until we prove so.
 * ndebard : Sept 17, 2002 */
#if 0
struct COVEN_Dimension {
	int height;
	int width;
};

struct COVEN_IO {
	void *data;
	char *tag;
	int COVEN_Type;
	struct COVEN_Dimension dimension;
};

struct COVEN_Module {
	char *module_name;
	int slave_module; /* is this a slave module or master */
	int task_number; /* which task does this run under */
	char **arguments;
	char **next_arg;
	/*
	char **parameters;
	char **input_tags;
	char **output_tags;
	*/
	struct COVEN_Interface *interface;
	int program_number;
};
#endif
#endif /* __COVEN_STRUCTS_H */
