/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

#include "coven.h"
#include <stdlib.h>
#include <string.h>

int COVEN_Find_Attribute_Next_Number(struct COVEN_TPH *tph,
  struct COVEN_Attribute *ptr)
{
	int i = 0;
	char tmp[256];
	struct COVEN_Attribute *ptr2;

	/* first, see how many others of these we have */
	do {
	    i++;
	    sprintf(tmp, "%s.%d", ptr->tag, i);
	    ptr2 = COVEN_Attribute_Table_Find_Attrib(tph->at, tmp);
	} while(ptr2 != NULL);
	/* now i equals the number we can move to */

	return i;
}

int COVEN_Modify_Attribute_Name(struct COVEN_TPH *tph, 
  struct COVEN_Attribute *ptr)
{
	int i = 0;
	char tmp[256];

	i = COVEN_Find_Attribute_Next_Number(tph, ptr);
	/* change name of ptr->tag to a new one */
	sprintf(tmp, "%s.%d", ptr->tag, i);
	COVEN_free(ptr->tag);
	ptr->tag = (char*)COVEN_malloc(strlen(tmp)+1);
	strcpy(ptr->tag, tmp);
	//printf("TAG = '%s', PTRTAG = '%s'\n", tmp, ptr->tag); fflush(stdout);

	return i+1;
}

void COVEN_Copy_Attribute_Table(struct COVEN_TPH *oldtph, 
  struct COVEN_TPH *newtph)
{
	struct COVEN_Attribute *cur, *ptr;
	int i;
	char tmp[256];
	int needToWrite = FALSE;
	
	cur = oldtph->at->head;
	while(cur != NULL) {
 	    needToWrite = FALSE;
	    /* try and see if we already have an attribute of this name */
	    ptr = COVEN_Attribute_Table_Find_Attrib(newtph->at, cur->tag);
#if 0
		if(ptr != NULL) {
			/* this sucks */
			if(ptr->attribute_type == cur->attribute_type) { 
				/* only do this if they are the same type */
				switch(cur->attribute_type) {
					case COVEN_CHAR:
						if((*(char*)(cur->data)) != (*(char*)(ptr->data))) {
							i = COVEN_Modify_Attribute_Name(newtph, ptr);
							snprintf(tmp, 256, "%s.%d", cur->tag, i);
							needToWrite = TRUE;
						}
						break;
					case COVEN_UCHAR:
						if((*(unsigned char*)(cur->data)) != (*(unsigned char*)(ptr->data))) {
							i = COVEN_Modify_Attribute_Name(newtph, ptr);
							snprintf(tmp, 256, "%s.%d", cur->tag, i);
							needToWrite = TRUE;
						}
						break;
					case COVEN_SHORT:
						if((*(short*)(cur->data)) != (*(short*)(ptr->data))) {
							i = COVEN_Modify_Attribute_Name(newtph, ptr);
							snprintf(tmp, 256, "%s.%d", cur->tag, i);
							needToWrite = TRUE;
						}
						break;
					case COVEN_USHORT:
						if((*(unsigned short*)(cur->data)) != (*(unsigned short*)(ptr->data))) {
							i = COVEN_Modify_Attribute_Name(newtph, ptr);
							snprintf(tmp, 256, "%s.%d", cur->tag, i);
							needToWrite = TRUE;
						}
						break;
					case COVEN_INT:
						if((*(int*)(cur->data)) != (*(int*)(ptr->data))) {
							i = COVEN_Modify_Attribute_Name(newtph, ptr);
							snprintf(tmp, 256, "%s.%d", cur->tag, i);
							needToWrite = TRUE;
						}
						break;
					case COVEN_UINT:
						if((*(unsigned int*)(cur->data)) != (*(unsigned int*)(ptr->data))) {
							i = COVEN_Modify_Attribute_Name(newtph, ptr);
							snprintf(tmp, 256, "%s.%d", cur->tag, i);
							needToWrite = TRUE;
						}
						break;
					case COVEN_FLOAT:
						if((*(float*)(cur->data)) != (*(float*)(ptr->data))) {
							i = COVEN_Modify_Attribute_Name(newtph, ptr);
							snprintf(tmp, 256, "%s.%d", cur->tag, i);
							needToWrite = TRUE;
						}
						break;
					case COVEN_DOUBLE:
						if((*(double*)(cur->data)) != (*(double*)(ptr->data))) {
							i = COVEN_Modify_Attribute_Name(newtph, ptr);
							snprintf(tmp, 256, "%s.%d", cur->tag, i);
							needToWrite = TRUE;
						}
						break;
					case COVEN_STRING:
						if(strcmp((unsigned char*)cur->data, (unsigned char*)ptr->data)) {
							i = COVEN_Modify_Attribute_Name(newtph, ptr);
							snprintf(tmp, 256, "%s.%d", cur->tag, i);
							needToWrite = TRUE;
						}
						break;
				}
			}
			else {
				i = COVEN_Find_Attribute_Next_Number(newtph, cur);
				if(i != 1)
					sprintf(tmp, "%s.%d", cur->tag, i);
				else
					strncpy(tmp, cur->tag, 256);
				needToWrite = TRUE;
			}
		}
#endif
	    //else {
	    if(ptr == NULL) {
		i = COVEN_Find_Attribute_Next_Number(newtph, cur);
		if(i != 1)
		    sprintf(tmp, "%s.%d", cur->tag, i);
		else
		    strncpy(tmp, cur->tag, 256);
		needToWrite = TRUE;
	    }
	    /* copy this attribute to the new TPH */
	    if(needToWrite == TRUE) { 
		//ptr = COVEN_New_Attribute(cur->attribute_type, cur->tag, cur->data);
		ptr = COVEN_New_Attribute(cur->attribute_type, tmp, cur->data);
		COVEN_Attribute_Table_Force_Attrib(newtph->at, ptr);
	    }
	    cur = cur->next;
	}
}

void COVEN_Change_Attrib_Tag(struct COVEN_Attribute_Table *at,
  char *oldtag, char *newtag)
{
	struct COVEN_Attribute *cur = at->head;

	while(cur != NULL) {
	    //printf("*** (cur = %s) LOOKING FOR TAG '%s' -> change to '%s'.\n", cur->tag, oldtag, newtag); fflush(stdout);
	    if(!strcmp(cur->tag, oldtag)) {
		/* found it, now change it over */
		COVEN_free(cur->tag);
		cur->tag = (char*)COVEN_malloc(strlen(newtag)+1);
		strcpy(cur->tag, newtag);
		//printf("*** FOUND IT, now it's = %s\n", cur->tag); fflush(stdout);
	    }
	    cur = cur->next;
	}
}

struct COVEN_Attribute *COVEN_New_Attribute(int type, char *tag,
  void *data)
{
	struct COVEN_Attribute *new_at = (struct COVEN_Attribute*)COVEN_malloc(
	  sizeof(struct COVEN_Attribute));

	new_at->attribute_type = type;
	COVEN_debug(DMARSHALL,"New_Attribute: mallocing tag.\n");
	new_at->tag = (char*)COVEN_malloc(strlen(tag)+1);
	strcpy(new_at->tag, tag);
	COVEN_debug(DMARSHALL,"New_Attribute: malloc data.\n");
	switch(new_at->attribute_type) {
	    case COVEN_CHAR:
		new_at->data = (char*)COVEN_malloc(sizeof(char));
		if(data != NULL) *((char*)(new_at->data)) = *(char*)data;
		break;
	    case COVEN_UCHAR:
		new_at->data = (unsigned char*)COVEN_malloc(sizeof(unsigned char));
		if(data != NULL) 
		    *((unsigned char*)(new_at->data)) = *(unsigned char*)data;
		break;
	    case COVEN_SHORT:
		new_at->data = (short*)COVEN_malloc(sizeof(short));
		if(data != NULL) *((short*)(new_at->data)) = *(short*)data;
		break;
	    case COVEN_USHORT:
		new_at->data = (unsigned short*)COVEN_malloc(sizeof(unsigned short));
		if(data != NULL) 
		    *((unsigned short*)(new_at->data)) = *(unsigned short*)data;
		break;
	    case COVEN_INT:
		new_at->data = (int*)COVEN_malloc(sizeof(int));
		if(data != NULL) 
		    *((int*)(new_at->data)) = *(int*)data;
		break;
	    case COVEN_UINT:
		new_at->data = (unsigned int*)COVEN_malloc(sizeof(unsigned int));
		if(data != NULL) 
		    *((unsigned int*)(new_at->data)) = *(unsigned int*)data;
		break;
	    case COVEN_FLOAT:
		new_at->data = (float*)COVEN_malloc(sizeof(float));
		if(data != NULL) 
		    *((float*)(new_at->data)) = *(float*)data;
		break;
	    case COVEN_DOUBLE:
		new_at->data = (double*)COVEN_malloc(sizeof(double));
		if(data != NULL) 
		    *((double*)(new_at->data)) = *(double*)data;
		break;
	    case COVEN_STRING:
		new_at->data = (char*)COVEN_malloc((strlen((char*)data)+1)*sizeof(
		  char));
		strcpy((char*)new_at->data,(char*)data);
		break;
	    default:
		COVEN_Error("Corrupt attribute type");
		break;
	}
	new_at->next = NULL;

	return new_at;
}

void COVEN_Print_Attribute(struct COVEN_Attribute *ptr)
{
	COVEN_debug(DTPH,"type = %d", ptr->attribute_type); 
	COVEN_debug(DTPH,"\ttag = \"%s\"", ptr->tag); 
	switch(ptr->attribute_type) {
	    case COVEN_CHAR:
		COVEN_debug(DTPH,"\tdata = %c\n", *((char*)(ptr->data)));
		break;
	    case COVEN_UCHAR:
		COVEN_debug(DTPH,"\tdata = %c\n", *((unsigned char*)(ptr->data)));
		break;
	    case COVEN_SHORT:
		COVEN_debug(DTPH,"\tdata = %d\n", *((short*)(ptr->data)));
		break;
	    case COVEN_USHORT:
		COVEN_debug(DTPH,"\tdata = %d\n", *((unsigned short*)(ptr->data)));
		break;
	    case COVEN_INT:
		COVEN_debug(DTPH,"\tdata = %d\n", *((int*)(ptr->data)));
		break;
	    case COVEN_UINT:
		COVEN_debug(DTPH,"\tdata = %d\n", *((unsigned int*)(ptr->data)));
		break;
	    case COVEN_FLOAT:
		COVEN_debug(DTPH,"\tdata = %f\n", *((float*)(ptr->data)));
		break;
  	    case COVEN_DOUBLE:
		COVEN_debug(DTPH,"\tdata = %g\n", *((double*)(ptr->data)));
		break;
	    case COVEN_STRING:
		COVEN_debug(DTPH,"\tdata = '%s'\n", ((unsigned char*)(ptr->data)));
		break;
	    default:
		COVEN_Error("Corrupt attribute type");
		break;
	}
}

void COVEN_Print_Attribute_Table(struct COVEN_Attribute_Table *at)
{
	struct COVEN_Attribute *ptr = at->head;
	COVEN_debug(DTPH,"[ATTRIBUTE TABLE]\n");
	while(ptr != NULL) {
	    COVEN_Print_Attribute(ptr); 
	    ptr = ptr->next;
	}
}

void *COVEN_Attribute_Table_Put_Attrib_Data(struct COVEN_Attribute_Table *at,
  char *tag, int type, int size)
{
	struct COVEN_Attribute *new_at;

	if(type == COVEN_STRING) {
	    char *tmp = (char*)COVEN_malloc(size*sizeof(char));
	    memset(tmp, '.', size-1);
	    tmp[size] = '\0';
	    new_at = COVEN_New_Attribute(type, tag, tmp);
	} 
	else {
	    new_at = COVEN_New_Attribute(type, tag, NULL);
	}

	COVEN_Attribute_Table_Force_Attrib(at, new_at);

	new_at = COVEN_Attribute_Table_Find_Attrib(at, tag);
	if(new_at == NULL) { /* umm, what the hell? */
	    char str[128];
	    sprintf(str,"Expected attribute '%s' not found, fatal error!", tag);
	    COVEN_Error(str);
	}

	/* we're gonna go ahead and zero out these values */
	switch(type) {
	    case COVEN_CHAR:
		*((char*)(new_at->data)) = 0;
	    	break;
	    case COVEN_UCHAR:
		*((unsigned char*)(new_at->data)) = 0;
	    	break;
	    case COVEN_SHORT:
		*((short *)(new_at->data)) = -1;
	    	break;
	    case COVEN_USHORT:
		*((unsigned short*)(new_at->data)) = 0;
	    	break;
	    case COVEN_INT:
		*((int*)(new_at->data)) = -1;
	    	break;
	    case COVEN_UINT:
		*((unsigned int*)(new_at->data)) = 0;
	    	break;
	    case COVEN_FLOAT:
		*((float*)(new_at->data)) = -1.0;
	    	break;
	    case COVEN_DOUBLE:
		*((double*)(new_at->data)) = -1.0;
	    	break;
	    /* don't do anything for string, at least not now */
	}
	
	return new_at->data;
}

void *COVEN_Attribute_Table_Get_Attrib_Data(struct COVEN_Attribute_Table *at,
  char *tag)
{
	struct COVEN_Attribute *attrib;
	
	COVEN_debug(DARGS, "ATTRIB_TABLE_GET_ATTRIB_DATA\n");
	attrib = COVEN_Attribute_Table_Find_Attrib(at, tag);
	if(attrib == NULL) { 
	    /* we didn't find an attrib you we expected to be there, that's 
	     * really bad */
	    char str[128];
	    sprintf(str,"Expected attribute '%s' not found, fatal error!", tag);
	    COVEN_Error(str);
	}
	return attrib->data;
}

void *COVEN_Attribute_Table_Get_Or_Create_Attrib_Data(
  struct COVEN_Attribute_Table *at, char *tag, int type, int size)
{
   	struct COVEN_Attribute *attrib;

	attrib = COVEN_Attribute_Table_Find_Attrib(at, tag);
	/* if we already have it created, just return it - fast */
	if(attrib != NULL) return attrib->data;

	/* else, we gotta create it - make this function create it */
	return COVEN_Attribute_Table_Put_Attrib_Data(at, tag, type, size);
}

int COVEN_Attribute_Table_Append_Attrib(struct COVEN_Attribute_Table *at, 
  struct COVEN_Attribute *new_at) 
{
	struct COVEN_Attribute *cur;

	if(at->head == NULL) { /* no head yet */
	    at->head = new_at;
	    at->tail = new_at;
	    return 1;
	}
	else {
	    cur = at->head;
	    /* run to the end */
	    while(cur->next != NULL) {
		/* No 2 attributes can have the same tag */
		if(!(strcmp(cur->tag, new_at->tag))) {
	 	    return 0;
		}
		cur = cur->next;
	    }
	    cur->next = new_at;
	    /* and now this is the tail */
	    at->tail = new_at;
	    return 1;
	}
}

int COVEN_Attribute_Table_Force_Attrib(struct COVEN_Attribute_Table *at,
  struct COVEN_Attribute *new_at)
{
	struct COVEN_Attribute *cur_at;

	cur_at = COVEN_Attribute_Table_Find_Attrib(at, new_at->tag);
	if(cur_at != NULL) {
	    COVEN_Attribute_Table_Delete_Attrib(at, cur_at);
	}
	COVEN_Attribute_Table_Append_Attrib(at, new_at);
	/* return 1 if we had to delete the current attribute before putting 
	 * the new one in */
	if(cur_at) return 1;
	else return 0;
}

int COVEN_Attribute_Table_Delete_Attrib(struct COVEN_Attribute_Table *at,
  struct COVEN_Attribute *find_me)
{
	struct COVEN_Attribute *cur = at->head;
	struct COVEN_Attribute *last = NULL;

	while(cur != NULL) {
	    /* found it! */
	    if(!(strcmp(cur->tag, find_me->tag))) {
		if(last) {
		    last->next = cur->next;
		    if(at->tail == cur) at->tail = last;
		}
		else { /* it was the 1st element */
		    at->head = cur->next;
		}
		/* now dealloc cur */
		COVEN_free(cur->tag);
		COVEN_free(cur->data);
		COVEN_free(cur);
		return 1;
	    }
	    last = cur;
	    cur = cur->next;
	}
	/* didn't find it */
	return 0;
}

struct COVEN_Attribute *COVEN_Attribute_Table_Find_Attrib(
  struct COVEN_Attribute_Table *at, char *find_tag)
{
	struct COVEN_Attribute *cur = at->head;

	while(cur != NULL) {
	    if(!strcmp(cur->tag, find_tag)) return cur;
	    cur = cur->next;
	}
	/* didn't find it */
	return NULL;
}

void COVEN_Free_Attribute_Table(struct COVEN_Attribute_Table *at)
{
	struct COVEN_Attribute *cur = at->head;
	struct COVEN_Attribute *tmp;

	COVEN_debug(DMARSHALL,"** Free_AT: starting..\n");
	
	while(cur != NULL) {
	    COVEN_debug(DMARSHALL,"** Free_AT: freeing it all...\n");
	    tmp = cur->next;
	    /* dealloc cur */
	    COVEN_free(cur->tag);
	    COVEN_free(cur->data);
	    COVEN_free(cur);
	    cur = tmp;
	}
	COVEN_debug(DMARSHALL,"** Free_AT: ending..\n");
}
