/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */


// Copyright (c) 2000 Walt Ligon, Clemson University, Scyld Computing Inc.
// All rights reserved.  This software is licensed under GPL.
// Author: Walt Ligon
// Date: Summer 2000

// $Header: /projects/cvsroot/coven-engine/src/coven_datatype.c,v 1.2 2003/01/28 14:44:11 ndebard Exp $
// $Log: coven_datatype.c,v $
// Revision 1.2  2003/01/28 14:44:11  ndebard
// GPLed it all.
//
// Revision 1.1.1.1  2003/01/23 12:18:18  ndebard
// Imported source - from Coven project, reorganization
//
// Revision 1.4  2002/11/13 03:20:03  ndebard
// The script now doesn't require you have a working directory and it also
// runs a make clean before starting.  I fixed a warning with a missing
// prototype as well.  Took out some annoying printfs that weren't using
// gossip.
//
// Revision 1.3  2002/10/28 18:36:12  ndebard
// The replace buffer stuff works now.  There seems to be a memory leak
// that needs to be found / fixed, however.
//
// Revision 1.2  2002/10/10 13:59:25  ndebard
// Rolled types2.h into types.h
//
// Revision 1.1  2002/10/10 13:40:45  ndebard
// Placing Walt's type stuff into the src place.
//
// Revision 1.2  2002/10/04 21:29:06  walt
// *** empty log message ***
//
// Revision 1.1  2002/09/10 18:50:36  walt
// initial COVEN version
//
// Revision 1.2  2002/08/06 15:37:21  walt
// changed to COVEN types
//
// Revision 1.1  2002/05/30 21:57:01  walt
// source files for datatype and distribution processing
//
// Revision 1.5  2000/08/09 15:48:14  walt
// working version 0.02
//
//

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <coven_types.h>
#include <cint_datatype.h>
#include <coven_datatype.h>
#include <coven_prototypes.h>

#define COVEN_SUCCESS 0
#define COVEN_ERR_REQ -1

/* elementary reqs */

static struct CINT_Type CINT_CHAR =
		{0, 1, 0, 1, 1, 0, 1, 1, 1, NULL, NULL};
COVEN_Type COVEN_CHAR = &CINT_CHAR;

static struct CINT_Type CINT_SHORT =
		{0, 1, 0, 1, 2, 0, 2, 1, 1, NULL, NULL};
COVEN_Type COVEN_SHORT = &CINT_SHORT;

static struct CINT_Type CINT_INT =
		{0, 1, 0, 1, 4, 0, 4, 1, 1, NULL, NULL};
COVEN_Type COVEN_INT = &CINT_INT;

static struct CINT_Type CINT_LONG =
		{0, 1, 0, 1, 4, 0, 4, 1, 1, NULL, NULL};
COVEN_Type COVEN_LONG = &CINT_LONG;

static struct CINT_Type CINT_UNSIGNED_CHAR =
		{0, 1, 0, 1, 1, 0, 1, 1, 1, NULL, NULL};
COVEN_Type COVEN_UNSIGNED_CHAR = &CINT_UNSIGNED_CHAR;

static struct CINT_Type CINT_UNSIGNED_SHORT =
		{0, 1, 0, 1, 2, 0, 2, 1, 1, NULL, NULL};
COVEN_Type COVEN_UNSIGNED_SHORT = &CINT_UNSIGNED_SHORT;

static struct CINT_Type CINT_UNSIGNED =
		{0, 1, 0, 1, 4, 0, 4, 1, 1, NULL, NULL};
COVEN_Type COVEN_UNSIGNED = &CINT_UNSIGNED;

static struct CINT_Type CINT_UNSIGNED_LONG =
		{0, 1, 0, 1, 4, 0, 4, 1, 1, NULL, NULL};
COVEN_Type COVEN_UNSIGNED_LONG = &CINT_UNSIGNED_LONG;

static struct CINT_Type CINT_FLOAT =
		{0, 1, 0, 1, 4, 0, 4, 1, 1, NULL, NULL};
COVEN_Type COVEN_FLOAT = &CINT_FLOAT;

static struct CINT_Type CINT_DOUBLE =
		{0, 1, 0, 1, 8, 0, 8, 1, 1, NULL, NULL};
COVEN_Type COVEN_DOUBLE = &CINT_DOUBLE;

static struct CINT_Type CINT_LONG_DOUBLE =
		{0, 1, 0, 1, 8, 0, 8, 1, 1, NULL, NULL};
COVEN_Type COVEN_LONG_DOUBLE = &CINT_LONG_DOUBLE;

static struct CINT_Type CINT_BYTE =
		{0, 1, 0, 1, 1, 0, 1, 1, 1, NULL, NULL};
COVEN_Type COVEN_BYTE = &CINT_BYTE;

static struct CINT_Type CINT_PACKED =
		{0, 1, 0, 1, 1, 0, 1, 1, 1, NULL, NULL};
COVEN_Type COVEN_PACKED = &CINT_PACKED;

int COVEN_Type_extent(COVEN_Type datatype, COVEN_size *extent);

int COVEN_Type_contiguous(COVEN_count32 num_blocks, COVEN_Type oldtype,
		COVEN_Type *newtype) 
{
	return COVEN_Type_hvector(1, num_blocks, 0, oldtype, newtype);
}

int COVEN_Type_vector(COVEN_count num_blocks, COVEN_size blocklength,
		COVEN_size stride, COVEN_Type oldtype, COVEN_Type *newtype)
{
	COVEN_size extent;
	if (oldtype == NULL)
		return COVEN_ERR_REQ;
	COVEN_Type_extent(oldtype, &extent);
	return COVEN_Type_hvector(num_blocks, blocklength, stride*extent,
			oldtype,newtype);
}

static int CINT_subreq(COVEN_offset offset, COVEN_count num_etype,
		COVEN_size stride, COVEN_count num_blocks,
		COVEN_Type oldtype, COVEN_size oldext, COVEN_Type *newtype)
{
	if (oldtype == NULL)
		return COVEN_ERR_REQ;
	(*newtype)->offset = offset;
	(*newtype)->num_etypes = num_etype;
	(*newtype)->stride = stride;
	(*newtype)->num_blocks = num_blocks;
	(*newtype)->ub = offset + ((num_blocks - 1) * stride) + (num_etype * oldext);
	(*newtype)->lb = offset;
	(*newtype)->aggregate_size = oldtype->aggregate_size * num_blocks * num_etype;
	(*newtype)->depth = oldtype->depth + 1;
	if (oldtype->aggregate_size != oldext)
		(*newtype)->num_contig_chunks = oldtype->num_contig_chunks *
			num_blocks * num_etype;
	else if (stride != num_etype * oldtype->aggregate_size &&
			stride != -(num_etype * oldtype->aggregate_size))
		(*newtype)->num_contig_chunks = num_blocks;
	else
		(*newtype)->num_contig_chunks = 1;
	(*newtype)->etype = oldtype;
	return COVEN_SUCCESS;
}

int COVEN_Type_hvector(COVEN_count num_blocks, COVEN_size blocklength,
		COVEN_size stride, COVEN_Type oldtype, COVEN_Type *newtype)
{
	COVEN_size oldext;
	if (oldtype == NULL)
		return COVEN_ERR_REQ;
	COVEN_Type_extent(oldtype, &oldext);
	*newtype = (CINT_Type *)COVEN_malloc(sizeof(struct CINT_Type));
	(*newtype)->stype = NULL;
	CINT_subreq(0, blocklength, stride, num_blocks, oldtype, oldext, newtype);
	if (stride < 0)
	{
		(*newtype)->lb = (num_blocks - 1) * stride;
	}
	return COVEN_SUCCESS;
}

int COVEN_Type_indexed(COVEN_count num_blocks, COVEN_size *blocklengths,
		COVEN_offset *displacements, COVEN_Type oldtype, COVEN_Type *newtype)
{
	CINT_Type *dt;
	COVEN_size oldext;
	*newtype = NULL;
	if (oldtype == NULL)
		return COVEN_ERR_REQ;
	COVEN_Type_extent(oldtype, &oldext);
	while (num_blocks--)
	{
		dt = *newtype;
		*newtype = (CINT_Type *)COVEN_malloc(sizeof(struct CINT_Type));
		(*newtype)->stype = dt;
		CINT_subreq(displacements[num_blocks]*oldext, blocklengths[num_blocks],
				0, 1, oldtype, oldext, newtype);
		if ((*newtype)->stype)
		{
			if ((*newtype)->lb > (*newtype)->stype->lb)
				(*newtype)->lb = (*newtype)->stype->lb;
			if ((*newtype)->ub < (*newtype)->stype->ub)
				(*newtype)->ub = (*newtype)->stype->ub;
			if ((*newtype)->depth < (*newtype)->stype->depth)
				(*newtype)->depth = (*newtype)->stype->depth;
			(*newtype)->aggregate_size = (*newtype)->aggregate_size +
				(*newtype)->stype->aggregate_size;
			(*newtype)->num_contig_chunks = (*newtype)->num_contig_chunks +
				(*newtype)->stype->num_contig_chunks;
			}
	}
	return COVEN_SUCCESS;
}

int COVEN_Type_hindexed(COVEN_count num_blocks, COVEN_size *blocklengths,
		COVEN_offset *displacements, COVEN_Type oldtype, COVEN_Type *newtype)
{
	CINT_Type *dt;
	COVEN_size oldext;
	*newtype = NULL;
	if (oldtype == NULL)
		return COVEN_ERR_REQ;
	COVEN_Type_extent(oldtype, &oldext);
	while (num_blocks--)
	{
		dt = *newtype;
		*newtype = (CINT_Type *)COVEN_malloc(sizeof(struct CINT_Type));
		(*newtype)->stype = dt;
		CINT_subreq(displacements[num_blocks], blocklengths[num_blocks], 0, 1,
				oldtype, oldext, newtype);
		if ((*newtype)->stype)
		{
			if ((*newtype)->lb > (*newtype)->stype->lb)
				(*newtype)->lb = (*newtype)->stype->lb;
			if ((*newtype)->ub < (*newtype)->stype->ub)
				(*newtype)->ub = (*newtype)->stype->ub;
			if ((*newtype)->depth < (*newtype)->stype->depth)
				(*newtype)->depth = (*newtype)->stype->depth;
			(*newtype)->aggregate_size = (*newtype)->aggregate_size +
				(*newtype)->stype->aggregate_size;
			(*newtype)->num_contig_chunks = (*newtype)->num_contig_chunks +
				(*newtype)->stype->num_contig_chunks;
			}
	}
	return COVEN_SUCCESS;
}

int COVEN_Type_struct(COVEN_count num_blocks, COVEN_size *blocklengths,
		COVEN_offset *displacements, COVEN_Type *oldtypes, COVEN_Type *newtype)
{
	CINT_Type *dt;
	COVEN_size oldext;
	*newtype = NULL;
	if (oldtypes == NULL)
		return COVEN_ERR_REQ;
	while (num_blocks--)
	{
		if (oldtypes[num_blocks] == NULL)
			return COVEN_ERR_REQ;
		COVEN_Type_extent(oldtypes[num_blocks], &oldext);
		dt = *newtype;
		*newtype = (CINT_Type *)COVEN_malloc(sizeof(struct CINT_Type));
		(*newtype)->stype = dt;
		CINT_subreq(displacements[num_blocks], blocklengths[num_blocks],
				0, 1, oldtypes[num_blocks], oldext, newtype);
		if ((*newtype)->stype)
		{
			if ((*newtype)->lb > (*newtype)->stype->lb)
				(*newtype)->lb = (*newtype)->stype->lb;
			if ((*newtype)->ub < (*newtype)->stype->ub)
				(*newtype)->ub = (*newtype)->stype->ub;
			if ((*newtype)->depth < (*newtype)->stype->depth)
				(*newtype)->depth = (*newtype)->stype->depth;
			(*newtype)->aggregate_size = (*newtype)->aggregate_size +
				(*newtype)->stype->aggregate_size;
			(*newtype)->num_contig_chunks = (*newtype)->num_contig_chunks +
				(*newtype)->stype->num_contig_chunks;
			}
	}
	return COVEN_SUCCESS;
}

int COVEN_Address(void* location, COVEN_offset *address)
{
	address = location;
	return COVEN_SUCCESS;
}

int COVEN_Type_extent(COVEN_Type datatype, COVEN_size *extent)
{
	if (datatype == NULL)
		return COVEN_ERR_REQ;
	*extent = datatype->ub - datatype->lb;
	return COVEN_SUCCESS;
}

int COVEN_Type_size(COVEN_Type datatype, COVEN_size *size)
{
	if (datatype == NULL)
		return COVEN_ERR_REQ;
	*size = datatype->aggregate_size;
	return COVEN_SUCCESS;
}

int COVEN_Type_lb(COVEN_Type datatype, COVEN_offset *displacement)
{
	if (datatype == NULL)
		return COVEN_ERR_REQ;
	*displacement = datatype->lb;
	return COVEN_SUCCESS;
}

int COVEN_Type_ub(COVEN_Type datatype, COVEN_offset *displacement)
{
	if (datatype == NULL)
		return COVEN_ERR_REQ;
	*displacement = datatype->ub;
	return COVEN_SUCCESS;
}

/*
int COVEN_Type_commit(COVEN_Type *datatype)
{
	if (datatype == NULL)
		return COVEN_ERR_REQ;
	return COVEN_SUCCESS;
}

int COVEN_Type_COVEN_free(COVEN_Type *datatype)
{
	if (datatype == NULL)
		return COVEN_ERR_REQ;
	return COVEN_SUCCESS;
}
int COVEN_Get_elements(COVEN_Status *status, COVEN_Type datatype,
		int *num_blocks)
{
	if (datatype == NULL)
		return COVEN_ERR_REQ;
	return COVEN_SUCCESS;
}
int COVEN_Pack_size(COVEN_count innum_blocks, COVEN_Type datatype,
		COVEN_Comm comm, COVEN_size *size) 
{
	if (datatype == NULL)
		return COVEN_ERR_REQ;
	*size = (datatype->aggregate_size * innum_blocks) + COVEN_BSEND_OVERHEAD;
	return COVEN_SUCCESS;
}
*/


