/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

/*
 * (C) 2002 Clemson University and The University of Chicago.
 *
 * See COPYING in top-level directory.
 */       

#include <stdlib.h>
#include <stdio.h>
#include <gossip.h>
#include <coven.h>
#include <coven_types.h>
#include <cint_datatype.h>

/* This function calls COVEN_Distribute for each contiguous chunk */
/* of the type.  COVEN_Distribute returns the number of bytes */
/* processed.  If this is less than the total bytes in the chunk */
/* this function returns otherwise it keeps processing until all */
/* chunks are done.  Returns the number of bytes processed.  It */
/* is assumed caller we retry if this is less than the total bytes */
/* in the type */

int CINT_Process_type(CINT_Type_state *type_state,
	COVEN_count tph_count,
	COVEN_TPH *tph,
	COVEN_Buffer buffer,
	COVEN_flag mode)
{
	COVEN_count target_tph;		/* TPH being gathered */
	COVEN_count current_tph;	/* TPH being accessed */
	void *buffer1;					/* TPH buffer for data copy */
	void *buffer2;					/* TPH buffer for data copy */
	void *temp_space = NULL;     /* temp copy of type state for size call */
	COVEN_boolean lvl_flag;      /* indicates level should be decremented */
	COVEN_offset  contig_offset; /* temp for offset of a contig region */
	COVEN_size    contig_size;   /* temp for size of a contig region */
	COVEN_size    retval;        /* return value from calls to distribute */
	COVEN_size    bytes_processed;
	COVEN_count32 segs_processed;
	COVEN_debug(DATATYPE_DEBUG,"CINT_Process_type\n");
	/* do very basic error checking here */
	if (!type_state)
	{
		gossip_lerr("CINT_Process_type: Bad CINT_Type_state!\n");
		return -1;
	}
	/* initialize some variables */
	segs_processed = 0;
	bytes_processed = 0;
	retval = 0;
	/* this section added for Coven - WBL */
	current_tph = 0;
	if (current_tph >= tph_count)
	{
		gossip_lerr("CINT_Process_type: no TPH's\n");
		return -1;
	}
	COVEN_debug(DATATYPE_DEBUG,"buffer addr: %x\n",
	  (unsigned int)buffer);
	COVEN_debug(DATATYPE_DEBUG,"buffer data addr (nate way): %x\n",
	  (unsigned int)COVEN_Buffer_Data(buffer));
	buffer1 = COVEN_Buffer_Data(buffer);
	COVEN_debug(DATATYPE_DEBUG,"buffer1 (which is buffer->data) "
	  "addr: %x\n", (unsigned int)buffer1);
	if (mode == COVEN_SCATTER)
	{
		/* create the first buffer */
		int target_size; /* size of new buffer in bytes */
		target_size = type_state->cur[0].ty->etype->aggregate_size *
				type_state->cur[0].ty->num_etypes;
		buffer2 = CINT_create_buffer(tph[current_tph],
						tph_count, current_tph, buffer->tag,
						buffer->type, target_size);
	}
	else /* mode == COVEN_GATHER */
	{
		/* check first buffer */
		buffer2 = CINT_get_buffer(tph[current_tph],
		  &tph_count, &target_tph, buffer->tag, buffer->type);
		COVEN_debug(DATATYPE_DEBUG,"GATHER for tph %d of %d\n",
				target_tph, tph_count);
		if (target_tph != 0)
		{
			COVEN_debug(DATATYPE_DEBUG,"Going into SEEK mode\n");
			mode = COVEN_SEEK;
		}
	}
	/* end of Coven specific code - WBL */
	/* we should be ready to begin */
	/* zero retval indicates everything flowing successfully */
	/* positive retval indicates a partial chunk was processed - so we */
	/* wait until later to retry */
	while(!retval)
	{
		if (type_state->cur[type_state->lvl].ty)
		{
		COVEN_debug(DATATYPE_DEBUG,"\tDo seq of %lld ne %d st %lld nb %d ",
				type_state->cur[type_state->lvl].ty->offset,
				type_state->cur[type_state->lvl].ty->num_etypes,
				type_state->cur[type_state->lvl].ty->stride,
				type_state->cur[type_state->lvl].ty->num_blocks);
		COVEN_debug(DATATYPE_DEBUG,"ub %lld lb %lld as %lld co %llu\n",
				type_state->cur[type_state->lvl].ty->ub,
				type_state->cur[type_state->lvl].ty->lb,
				type_state->cur[type_state->lvl].ty->aggregate_size,
				type_state->cur[type_state->lvl].chunk_offset);
		COVEN_debug(DATATYPE_DEBUG,"\t\tlvl %d el %d blk %d by %lld bo %lld\n",
				type_state->lvl,
				type_state->cur[type_state->lvl].el,
				type_state->cur[type_state->lvl].blk,
				type_state->bytes,
				type_state->buf_offset);
		}
		/* in seek mode do not transfer data, skip to next element */
		/* this is COVEN specific - WBL */
		if (mode == COVEN_SEEK)
		{
				goto return_from_level;
		}
		/* end COVEN specific - WBL */
		/* NULL type indicates packed data - handle directly */
		if (type_state->cur[type_state->lvl].ty == NULL)
		{
			COVEN_debug(DATATYPE_DEBUG,"\tnull type\n");
			contig_offset = type_state->cur[type_state->lvl].chunk_offset +
				type_state->bytes;
			contig_size = type_state->cur[type_state->lvl].maxel -
				type_state->bytes;
			lvl_flag = 1;
		}
		/* basic data type or contiguous data - handle directly */
		/* NULL etype indicates current type is packed bytes */
		/* current type is contiguous because its size equals ub minus lb */
		/* COVEN does not process level 0 as contiguous */
		else if ((type_state->cur[type_state->lvl].ty->etype == NULL ||
				type_state->cur[type_state->lvl].ty->aggregate_size ==
				(type_state->cur[type_state->lvl].tybase->ub -
					type_state->cur[type_state->lvl].tybase->lb)) &&
				type_state->cur[type_state->lvl].ty ==
				type_state->cur[type_state->lvl].tybase &&
				type_state->lvl != 0 /* added for Coven - WBL */)
		{
			COVEN_debug(DATATYPE_DEBUG,"\tbasic type or contiguous data\n");
			contig_offset = type_state->cur[type_state->lvl].chunk_offset +
				type_state->bytes;
			contig_size = (type_state->cur[type_state->lvl].maxel *
					type_state->cur[type_state->lvl].ty->aggregate_size) -
				type_state->bytes;
			lvl_flag = 1;
		}
		/* subtype is contiguous because its size equals its ub minus lb */
		else if (type_state->cur[type_state->lvl].ty->etype->aggregate_size ==
				(type_state->cur[type_state->lvl].ty->etype->ub -
				type_state->cur[type_state->lvl].ty->etype->lb))
		{
			COVEN_debug(DATATYPE_DEBUG,"\tsubtype is contiguous\n");
			contig_offset = type_state->cur[type_state->lvl].chunk_offset +
				(type_state->cur[type_state->lvl].el *
				(type_state->cur[type_state->lvl].tybase->ub -
				type_state->cur[type_state->lvl].tybase->lb)) +
				type_state->cur[type_state->lvl].ty->offset +
				(type_state->cur[type_state->lvl].ty->stride *
				type_state->cur[type_state->lvl].blk) + type_state->bytes;
			contig_size =
				(type_state->cur[type_state->lvl].ty->etype->aggregate_size *
				type_state->cur[type_state->lvl].ty->num_etypes) -
				type_state->bytes;
			lvl_flag = 0;
		}
		/* go to the next level and "recurse" */
		else
		{
			COVEN_debug(DATATYPE_DEBUG,"\tgoing to next level %d\n",
					type_state->lvl+1);
			type_state->cur[type_state->lvl+1].el = 0;
			type_state->cur[type_state->lvl+1].maxel =
				type_state->cur[type_state->lvl].ty->num_etypes;
			type_state->cur[type_state->lvl+1].ty =
				type_state->cur[type_state->lvl].ty->etype;
			type_state->cur[type_state->lvl+1].tybase =
				type_state->cur[type_state->lvl].ty->etype;
			type_state->cur[type_state->lvl+1].blk = 0;
			type_state->cur[type_state->lvl+1].chunk_offset =
				type_state->cur[type_state->lvl].chunk_offset +
					(type_state->cur[type_state->lvl].el *
					(type_state->cur[type_state->lvl].tybase->ub -
					type_state->cur[type_state->lvl].tybase->lb)) +
					type_state->cur[type_state->lvl].ty->offset +
					(type_state->cur[type_state->lvl].ty->stride *
					type_state->cur[type_state->lvl].blk);
			type_state->lvl++;
			continue;
		}
		/* we process the whole thing at once */
		retval = CINT_Distribute(contig_offset, contig_size, 
				&bytes_processed, &segs_processed, 
				buffer1, buffer2,
				mode);
		type_state->buf_offset += retval;
		/* see if we processed all of the bytes expected */
		if (retval != contig_size)
		{
			/* no so record the bytes processed */
			type_state->bytes += retval;
			break;
		}
		/* processed all bytes so continue on and return from level */
		if (lvl_flag)
		{
			type_state->lvl--;
		}

	return_from_level:
		COVEN_debug(DATATYPE_DEBUG,"\treturn from level %d\n",type_state->lvl);
		retval = 0;
		type_state->bytes = 0;
		if (type_state->lvl < 0)
		{
			/* we have processed the entire type */
			break;
		}

		/* go to the next block */
		COVEN_debug(DATATYPE_DEBUG,"\tgoing to next block\n");
		type_state->cur[type_state->lvl].blk++;
		if (type_state->cur[type_state->lvl].blk >=
				type_state->cur[type_state->lvl].ty->num_blocks)
		{
			/* that was the last block */
			type_state->cur[type_state->lvl].blk = 0;
			/* go to next item in sequence chain */
			COVEN_debug(DATATYPE_DEBUG,"\tgoing to next item in sequence chain\n");
			type_state->cur[type_state->lvl].ty =
				type_state->cur[type_state->lvl].ty->stype;
			if (type_state->cur[type_state->lvl].ty == NULL)
			{
				/* that was last item in sequence chain */
				type_state->cur[type_state->lvl].ty =
					type_state->cur[type_state->lvl].tybase;
				/* go to next element in block of level above */
				COVEN_debug(DATATYPE_DEBUG,
						"\tgoing to next element in block of level above\n");
				type_state->cur[type_state->lvl].el++;
				if (type_state->cur[type_state->lvl].el >=
						type_state->cur[type_state->lvl].maxel)
				{
					/* that was last element in block of level above */
					type_state->lvl--;
					/* go back up a level */
					goto return_from_level;
				}
			}
		}

		/* this section added for Coven - WBL */
		if (type_state->lvl == 0)
		{
			COVEN_debug(DATATYPE_DEBUG,"\tGOING TO NEXT LEVEL 0 ELEMENT\n");
			current_tph++;
			bytes_processed = 0;
			if (current_tph >= tph_count)
			{
				COVEN_debug(DATATYPE_DEBUG,"\tNo more TPHs\n");
				break;
			}
			if (mode == COVEN_SCATTER)
			{
				/* create next TPH */
				int target_size; /* size of new buffer in bytes */
				target_size = type_state->cur[0].ty->etype->aggregate_size *
						type_state->cur[0].ty->num_etypes;
				buffer2 = CINT_create_buffer(tph[current_tph],
								tph_count, current_tph, buffer->tag,
								buffer->type, target_size);
			}
			else if (mode == COVEN_SEEK)
			{
				COVEN_debug(DATATYPE_DEBUG,"\tCheck for target does %d = %d?\n",
						current_tph, target_tph);
				/* see if we have found the right TPH */
				if (current_tph == target_tph)
				{
					COVEN_debug(DATATYPE_DEBUG,"\t\tYES! Leaving SEEK mode\n");
					mode = COVEN_GATHER;
				}
			}
			else /* mode == COVEN_GATHER */
			{
				/* we only process one TPH in gather - we are done now */
				COVEN_debug(DATATYPE_DEBUG,"\tFinished with GATHER\n");
				break;
			}
		}
		/* end of Coven specific code - WBL */

	} /* this is the end of the while loop */
	COVEN_debug(DATATYPE_DEBUG,"\tdone\n");
	if (mode == CINT_CKSIZE)
	{
		/* restore type state */
		COVEN_free(temp_space);
	}
	return bytes_processed;
}

/* This function creates a type state and sets it up to begin */
/* processing a type */
struct CINT_Type_state *CINT_New_datatype_state (CINT_Type *datatype)
{
	struct CINT_Type_state *type_state;
	COVEN_count32 tydepth;
	COVEN_debug(DATATYPE_DEBUG,"CINT_New_datatype_state\n");
	if (!(type_state =
				(struct CINT_Type_state *)COVEN_malloc(sizeof(struct CINT_Type_state))))
	{
		gossip_lerr("CINT_New_datatype_state failed to malloc type_state !!!!!!!\n");
		return NULL;
	}
	type_state->lvl = 0;
	type_state->bytes = 0;
	type_state->buf_offset = 0;
	/* we assume null type is a contiguous byte range depth 1 */
	if (datatype)
	{
		tydepth = datatype->depth;
	}
	else
	{
		tydepth = 1;
	}
	if (!(type_state->cur = (struct CINT_typestack *)COVEN_malloc(tydepth *
			sizeof(struct CINT_typestack))))
	{
		gossip_lerr("CINT_New_datatype_state failed to malloc tystack !!!!!!\n");
		return NULL;
	}
		/* transfer one instance of type */
	type_state->cur[type_state->lvl].maxel = 1;
	type_state->cur[type_state->lvl].el = 0;
	type_state->cur[type_state->lvl].ty = datatype;
	type_state->cur[type_state->lvl].tybase = datatype;
	type_state->cur[type_state->lvl].blk = 0;
		/* transfer from inital file offset */
	type_state->cur[type_state->lvl].chunk_offset = 0;
	return type_state;
}

/* This function frees type state structures */
void CINT_Free_datatype_state (CINT_Type_state *type_state)
{
	COVEN_free(type_state->cur);
	COVEN_free(type_state);
}

/* combined function for a scatter */
int CINT_Scatter(CINT_Type *datatype,
	COVEN_count tph_count,
	COVEN_TPH *tph,
	COVEN_Buffer buffer)
{
	struct CINT_Type_state *tystate;
	int retval;

	tystate = CINT_New_datatype_state(datatype);
	retval = CINT_Process_type(tystate, tph_count, tph, buffer, COVEN_SCATTER);
	CINT_Free_datatype_state(tystate);
	return retval;
}

/* combined function for a gather */
int CINT_Gather(CINT_Type *datatype,
	COVEN_count tph_count,
	COVEN_TPH *tph,
	COVEN_Buffer buffer)
{
	struct CINT_Type_state *tystate;
	int retval;

	tystate = CINT_New_datatype_state(datatype);
	retval = CINT_Process_type(tystate, tph_count, tph, buffer, COVEN_GATHER);
	CINT_Free_datatype_state(tystate);
	return retval;
}

/* This function should return the byte displacement from the input argument
 * offset to the last byte in the segment processed regardless of
 * whether that byte is in the current distribution or not.
 * Inputs:
 * 	- offset and size are the contiguous region in logical file
 * 	space we are to process
 * 	- buffer1 is a pointer to the master TPH
 * 	- buffer2 is a pointer to the distributed TPH
 * 	- bytes is the position in the distributed TPH where data is to be
 * 	stored
 * 	- mode indicates COVEN_SCATTER or COVEN_GATHER
 * Outputs:
 * 	- returns number of bytes processed
 */
COVEN_size CINT_Distribute(COVEN_offset offset, COVEN_size size,
		COVEN_size *bytes, 
		COVEN_count32 *segs, 
		void *buffer1,
		void *buffer2,
		COVEN_flag mode)
{
	COVEN_debug(DATATYPE_DEBUG,"copy offset:%lld size:%lld bytes:%lld\n",
			offset, size, *bytes);
	COVEN_debug(DATATYPE_DEBUG,"\tbuf1:%x buf2:%x\n",
			(unsigned int)buffer1, (unsigned int)buffer2);
	if (mode == COVEN_SCATTER)
	{
		memcpy(buffer2+*bytes, buffer1+offset, size);
	}
	else /* mode == COVEN_GATHER */
	{
		memcpy(buffer1+offset, buffer2+*bytes, size);
	}
	(*bytes) += size;
	(*segs)++;
	return size;
}

/* this function takes a TPH, a tag, a size, and a type and creates
 * a buffer in the TPH of that size, type, and with that tag.  It also
 * hooks up the buffer into the TPH.
 * Additonally, it sets up the stack-based "scatter" information
 * which contains stats of what buffer this is out of the number of
 * bufs that are being created (scattered).
 * The code that calls this expects the buffer numbers to range
 * from 0->N-1.  We want to keep them as 1->N, so we increment that
 * here and are careful to decrement the value on this functions
 * complement. */
void* CINT_create_buffer(struct COVEN_TPH *tph,
  int num_bufs, int this_buf_num, char *tag, int ptype, int numbytes)
{
	struct COVEN_Buffer *buf;
	int single_el_size; /* size of a single element, in bytes */

	/* COVEN_New_Buffer expects the number of elements, not the number of
	 * bytes, so convert. */
	if(ptype < PSE_TYPE_OFFSET) 
	    single_el_size = COVEN_get_type_size(ptype);
	else
	    single_el_size = PSE_get_type_size(ptype);
	buf = COVEN_New_Buffer(tph, tag, numbytes / single_el_size, ptype);
	buf->seg_num = this_buf_num + 1;
	buf->num_segs = num_bufs;

	COVEN_Print_TPH(tph);
	return COVEN_Buffer_Data(buf);
}

void* CINT_get_buffer(struct COVEN_TPH *tph,
  int *num_bufs, int *this_buf_num, char *tag, int ptype)
{
	struct COVEN_Buffer *buf;

	buf = COVEN_Get_Buffer(tph, tag, ptype);
	*num_bufs = buf->num_segs;
	/* the buffer processing code assumes the buffers range from
	 * [0->N-1] but we stored them as [1->N].  So decrement
	 * the value here before returning. */
	*this_buf_num = (buf->seg_num) -1;

	return COVEN_Buffer_Data(buf);
}
