/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

#include "coven.h"
#include <stdlib.h>

extern int myMPIrank, myCOVENTaskID, myProcessorNumber, am_slave;
unsigned int total_alloced = 0;

void * COVEN_malloc(int num_bytes)
{
	int real_num_bytes = num_bytes + 2*sizeof(unsigned int);
	void *ptr;
	int mask = DMEMV;

	ptr = malloc(real_num_bytes);
	if(ptr == NULL) {
	    /* ERROR, we didn't get our memory! */
	   COVEN_Error("Malloc failed.\n");
	}
	((unsigned int *)ptr)[0] = COVEN_MAGIC;
	((unsigned int *)ptr)[1] = num_bytes;

	COVEN_debug(mask, "%x MEM TOT (%d) Malloced %d user, %d system bytes.  Magic = %d, "
	  "stored bytes = %d, sys addr = %x, user addr = %x\n", 
	  (unsigned int)((unsigned char*)ptr) + 2*sizeof(unsigned int),
	  total_alloced, num_bytes, real_num_bytes,
	  ((unsigned int *)ptr)[0], 
	  ((unsigned int *)ptr)[1], 
	  (unsigned int)ptr, 
	  (unsigned int)((unsigned char*)ptr) + 2*sizeof(unsigned int));

	total_alloced += num_bytes;
	return ((unsigned char*)ptr) + 2*sizeof(unsigned int);
}

void COVEN_free(void *ptr)
{
	unsigned int magic, num_bytes;
	void *sys_ptr;
	int mask = DMEMV;

	sys_ptr = (unsigned char*)ptr - 2*sizeof(unsigned int);
	magic = ((unsigned int*)sys_ptr)[0];
	num_bytes = ((unsigned int*)sys_ptr)[1];

	if(magic != COVEN_MAGIC) mask = DMEM;
	
	COVEN_debug(mask, "%x MEM TOT (%d) Free user addr %x, sys addr %x.  Magic = %d, "
	  " num bytes = %d.\n", (unsigned int)ptr, total_alloced,
	  (unsigned int)ptr, (unsigned int)sys_ptr, magic, num_bytes);
	
	total_alloced -= num_bytes;
	free(sys_ptr);

}

char * COVEN_Get_Debug_Prepend()
{	
	static char str[128];
	snprintf(str, 128, "[C(%c:%d),P:%d,M:%d]", 
	  am_slave ? 'S' : 'M',
	  myCOVENTaskID,
	  myProcessorNumber,
	  myMPIrank);
	return str;
}

void COVEN_Print_Program(struct COVEN_Program *pgm)
{
	COVEN_debug(DBND,"-- [ COVEN PROGRAM ] @START --\n");
	COVEN_debug(DBND,"  Number of master tasks: %d\n", pgm->num_master_tasks);
	COVEN_debug(DBND,"  Number of slave tasks: %d\n", pgm->num_slave_tasks);
	COVEN_Print_Consts(pgm->constants);
	COVEN_Print_Mods(pgm->modules);
	COVEN_Print_Mod_Calls(pgm->mod_calls);
	COVEN_debug(DBND,"-- [ COVEN PROGRAM ] @END --\n");
}

void COVEN_Print_Consts(struct COVEN_Consts *consts)
{
   	int i;

   	COVEN_debug(DBND,"  -- [ COVEN CONSTS ] @START --\n");
	COVEN_debug(DBND,"    Number of constants: %d\n", consts->num_consts);
	for(i=0; i<consts->num_consts; i++)
	    COVEN_Print_Const(&(consts->const_decls[i]));
   	COVEN_debug(DBND,"  -- [ COVEN CONSTS ] @END --\n");
}

void COVEN_Print_Const(struct COVEN_Const_Decl *cd)
{
   	COVEN_debug(DBND,"    CONST (type:%d) (GB:'%s') (STRV:'%s')\n",
	  cd->type, cd->global_binding, cd->str_value);
}

void COVEN_Print_Mods(struct COVEN_Mods *mods)
{
	int i;

	COVEN_debug(DBND,"  -- [ COVEN MOD(ule)S ] @START --\n");
	COVEN_debug(DBND,"    Number of modules: %d\n", mods->num_mod_decls);
	for(i=0; i<mods->num_mod_decls; i++)
	    COVEN_Print_Mod_Decl(&(mods->mod_decls[i]));
	COVEN_debug(DBND,"  -- [ COVEN MOD(ule)S ] @END --\n");
}

void COVEN_Print_Mod_Decl(struct COVEN_Mod_Decl *md)
{
	COVEN_debug(DBND,"    -- [ COVEN MODULE DECLARATION ] @START --\n");
	COVEN_debug(DBND,"      Shared object filename: %s\n", md->file_name);
	COVEN_debug(DBND,"      Module name: %s\n", md->module_name);
	COVEN_Print_Interface(md->interface);
	COVEN_debug(DBND,"    -- [ COVEN Module Declaration ] @END --\n");
}

void COVEN_Print_Interface(struct COVEN_Interface *iface)
{
   	int i;

   	COVEN_debug(DBND,"      -- [ COVEN INTERFACE ] @START --\n");
	if(iface == NULL || iface->iface_elems == NULL) {
	    COVEN_debug(DBND,"        * NULL * NULL * NULL * NULL *\n");
	}
	else {
	    COVEN_debug(DBND,"        Number of interface elements: %d\n",
	      iface->num_iface_elems);
	    for(i=0; i<iface->num_iface_elems; i++)
	    	COVEN_Print_Interface_Element(&(iface->iface_elems[i]));
	}
   	COVEN_debug(DBND,"      -- [ COVEN INTERFACE ] @END --\n");
}

void COVEN_Print_Interface_Element(struct COVEN_Interface_Element *iel)
{
   	COVEN_debug(DBND,"        IFACE_EL: (LB:'%s') (type:%d) (mode:%d)\n",
	  iel->local_binding, iel->type, iel->mode);
}

void COVEN_Print_Mod_Calls(struct COVEN_Mod_Calls *calls)
{
	int i;

   	COVEN_debug(DBND,"  -- [ COVEN MOD(ule) CALLS ] @START --\n");
	COVEN_debug(DBND,"    Number of module calls: %d\n", calls->num_mod_calls);
	for(i=0; i<calls->num_mod_calls; i++)
	    COVEN_Print_Mod_Call(&(calls->mod_calls[i]));
   	COVEN_debug(DBND,"  -- [ COVEN MOD(ule) CALLS ] @END --\n");
}

void COVEN_Print_Mod_Call(struct COVEN_Mod_Call *call)
{
   	COVEN_debug(DBND,"    -- [ COVEN MOD(ule) CALL ] @START --\n");
	COVEN_debug(DBND,"      Module name: %s\n", call->module_name);
	COVEN_debug(DBND,"      Slave (1) or master (0)? : %d\n", call->slave_module);
	COVEN_debug(DBND,"      \"Thread\" number: %d\n", call->task_number);
	COVEN_debug(DBND,"      Program number???: %d\n", call->program_number);
	COVEN_Print_Binding_Table(call->btable);
	COVEN_Print_Interface(call->interface);
   	COVEN_debug(DBND,"    -- [ COVEN MOD(ule) CALL ] @END --\n");
}

void COVEN_Print_Binding_Table(struct COVEN_Mod_Binding_Table *btable)
{
   	int i;

   	COVEN_debug(DBND,"      -- [ COVEN BINDING TABLE ] @START --\n");
	COVEN_debug(DBND,"        Number of binding table elements: %d\n",
	  btable->num_bindings);
	COVEN_debug(DBND,"        Index to look for next binding: %d\n",
	  btable->try_here);
	for(i=0; i<btable->num_bindings; i++)
	    COVEN_Print_Binding(&(btable->bindings[i]));
   	COVEN_debug(DBND,"      -- [ COVEN BINDING TABLE ] @END --\n");
}

void COVEN_Print_Binding(struct COVEN_Mod_Binding *binding)
{
   	COVEN_debug(DBND,"        LB:'%s' -> GB:'%s'\n",
	  binding->local_binding, binding->global_binding);
}
