/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

#include "coven.h"
#include <stdlib.h>
#include <string.h>

void COVEN_Marshall_TPH(struct COVEN_TPH *tph, 
  void **marshalled_TPH, int *marshalled_TPH_size)
{
	struct COVEN_Buffer_List_El *bufel = tph->head;
	struct COVEN_Attribute *cur = tph->at->head;
	COVEN_Profile_Stamp *cur_ps = tph->pt->head;
	unsigned int numbufs = 0;
	unsigned int num_attribs = 0;
	unsigned int num_stamps = 0;
	unsigned int bytes = 0;
	unsigned int offset = 0;
	
	/* we need a little extra to store the number of palantir stamps */
	bytes += sizeof(unsigned int);
	while(cur_ps) {
	    bytes += sizeof(unsigned int)*6 + sizeof(_stamp_type) +
	      128*sizeof(unsigned char) + sizeof(struct timeval);
	    cur_ps = cur_ps->next;
	    num_stamps ++;
	}
	
	/* we need a little extra to store the number of attributes */
	bytes += sizeof(unsigned int);
	/* now get the num of bytes for the attributes */
	while(cur) {
	    /* space for the attribute type and tag length */
	    bytes += 2*sizeof(unsigned int);
	    /* the bytes for the tag */
	    bytes += ((strlen(cur->tag)+1) * sizeof(char));
	    /* now the void * data, ooo, hard to figure out */
	    switch(cur->attribute_type) {
		case COVEN_CHAR: bytes += sizeof(char); break;
		case COVEN_UCHAR: bytes += sizeof(unsigned char); break;
		case COVEN_SHORT: bytes += sizeof(short); break;
		case COVEN_USHORT: bytes += sizeof(unsigned short); break;
		case COVEN_INT: bytes += sizeof(int); break;
		case COVEN_UINT: bytes += sizeof(unsigned int); break;
		case COVEN_FLOAT: bytes += sizeof(float); break;
		case COVEN_DOUBLE: bytes += sizeof(double); break;
		case COVEN_STRING: 
		    bytes += sizeof(unsigned int)+(strlen(cur->data)+1)*
		      sizeof(unsigned char); 
		    break;
		default: COVEN_Error("Marshall: Corrupt type");
	    }
	    cur = cur->next;
	    num_attribs++;
	}

	/* we need a little extra to store the number of buffers */
	bytes += sizeof(unsigned int);

	/* run through real fast and see how much contiguous memory we need
	 * to alloc */
	while(bufel) {
	    int size;

	    if(bufel->buffer->type < PSE_TYPE_OFFSET) 
	    	size = COVEN_get_type_size(bufel->buffer->type);
	    else 
	       	size = PSE_get_type_size(bufel->buffer->type);

	    numbufs++;
	    bytes += 5*sizeof(unsigned int)+1+strlen(bufel->buffer->tag)+
	      (bufel->buffer->size * size);
	    bufel = bufel->next;
	}

	/* reposition pointer to the head element */
	bufel = tph->head;

	/* here's the contiguous memory */
	*marshalled_TPH = (void*)COVEN_malloc(bytes); 
	/* setup the size pointer for the caller */
	*marshalled_TPH_size = bytes; 

	((unsigned int*)*marshalled_TPH)[0] = num_stamps;
	((unsigned int*)*marshalled_TPH)[1] = num_attribs;
	offset = 2*sizeof(unsigned int);

	cur_ps = tph->pt->head;
	while(cur_ps) {
	    ((unsigned int*)(*marshalled_TPH+offset))[0] = 
	      cur_ps->processor_number;
	    ((unsigned int*)(*marshalled_TPH+offset))[1] = 
	      cur_ps->mpirank;
	    ((unsigned int*)(*marshalled_TPH+offset))[2] = 
	      cur_ps->coven_task_id;
	    ((unsigned int*)(*marshalled_TPH+offset))[3] = 
	      cur_ps->TPH_num_tags;
	    ((unsigned int*)(*marshalled_TPH+offset))[4] = 
	      cur_ps->TPH_num_total_bytes;
	    offset += 5*sizeof(unsigned int);
	    ((_stamp_type*)(*marshalled_TPH+offset))[0] = cur_ps->stamp_type;
	    offset += sizeof(_stamp_type);
	    ((unsigned int*)(*marshalled_TPH+offset))[0] = cur_ps->special_int;
	    offset += sizeof(unsigned int);
	    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[0],
	      cur_ps->special_string, sizeof(unsigned char)*128);
	    offset += 128*sizeof(unsigned char);
	    ((struct timeval*)(*marshalled_TPH+offset))[0] = cur_ps->timestamp;
	    offset += sizeof(struct timeval);
	    cur_ps = cur_ps->next;
	}
	
	cur = tph->at->head;
	while(cur) {
	    /* put the tag length in there */
	    ((unsigned int*)(*marshalled_TPH+offset))[0] = strlen(cur->tag);
	    /* put the type in there */
	    ((unsigned int*)(*marshalled_TPH+offset))[1] = cur->attribute_type;
	    /* copy the tag in there */
	    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[2*
	      sizeof(unsigned int)],
	      cur->tag, (strlen(cur->tag)+1)*sizeof(unsigned char));
	    offset += 2*sizeof(unsigned int)+1+strlen(cur->tag);
	    switch(cur->attribute_type) {
		case COVEN_CHAR:
		    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[0],
		      cur->data, sizeof(char));
		    offset += sizeof(char);
		    break;
		case COVEN_UCHAR: 
		    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[0],
		      cur->data, sizeof(unsigned char));
		    offset += sizeof(unsigned char);
		    break;
		case COVEN_SHORT: 
		    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[0],
		      cur->data, sizeof(short));
		    offset += sizeof(short);
	 	    break;
		case COVEN_USHORT:
		    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[0],
		    cur->data, sizeof(unsigned short));
		    offset += sizeof(unsigned short);
		    break;
		case COVEN_INT:
		    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[0],
		    cur->data, sizeof(int));
		    offset += sizeof(int);
		    break;
		case COVEN_UINT:
		    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[0],
		    cur->data, sizeof(unsigned int));
		    offset += sizeof(unsigned int);
		    break;
		case COVEN_FLOAT: 
		    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[0],
		    cur->data, sizeof(float));
		    offset += sizeof(float);
		    break;
	        case COVEN_DOUBLE:
		    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[0],
		    cur->data, sizeof(double));
		    offset += sizeof(double);
		    break;
		case COVEN_STRING: 
		    /* but for strings we have to know the length of the 
		     * string */
		    ((unsigned int*)(*marshalled_TPH+offset))[0] = 
		      strlen(cur->data);
		    offset += sizeof(unsigned int);
		    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[0],
		    cur->data, (strlen(cur->data)+1)*sizeof(unsigned char));
		    offset += (strlen(cur->data)+1)*sizeof(unsigned char);
		    break;
	    }
	    cur = cur->next;
	}
	
	((unsigned int*)(*marshalled_TPH+offset))[0] = numbufs;
	offset += sizeof(unsigned int);

	/* loop through the buffers, marshalling them up as we go */
	while(bufel) {
	    int size;

	    if(bufel->buffer->type < PSE_TYPE_OFFSET) 
	    	size = COVEN_get_type_size(bufel->buffer->type);
	    else 
	       	size = PSE_get_type_size(bufel->buffer->type);
		
	    /* marshall tag */
	    ((unsigned int*)(*marshalled_TPH+offset))[0] = 
	      strlen(bufel->buffer->tag);
	    ((unsigned int*)(*marshalled_TPH+offset))[1] = 
	      bufel->buffer->type;
	    ((unsigned int*)(*marshalled_TPH+offset))[2] =
	      bufel->buffer->size;
	    ((unsigned int*)(*marshalled_TPH+offset))[3] =
	      bufel->buffer->seg_num;
	    ((unsigned int*)(*marshalled_TPH+offset))[4] =
	      bufel->buffer->num_segs;
	    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[5*
	      sizeof(unsigned int)],
	      bufel->buffer->tag, strlen(bufel->buffer->tag)+1);
	    memcpy(&((unsigned char*)(*marshalled_TPH+offset))[5*
	      sizeof(unsigned int)+
	      strlen(bufel->buffer->tag)+1], COVEN_Buffer_Data(bufel->buffer),
	      bufel->buffer->size * size);
	    offset += 5*sizeof(unsigned int)+1+strlen(bufel->buffer->tag)+
	      (bufel->buffer->size * size);
	    bufel = bufel->next;
	}
}

struct COVEN_TPH * COVEN_Unmarshall_TPH(void *marshalled_TPH)
{
	struct COVEN_TPH *tph;
	struct COVEN_Buffer *buf;
	struct COVEN_Attribute *new_at;
	COVEN_Profile_Stamp *new_ps;
	char *tag;
	unsigned int psize, ptype, seg_num, num_segs,
	  taglen, numbufs, num_stamps, num_attribs, 
	  i, offset, atype, strlength;
	void *adata = NULL;

	COVEN_debug(DMARSHALL," - unmarshall TPH starting\n");
	tph = COVEN_New_TPH();
	
	num_stamps = ((unsigned int*)marshalled_TPH)[0];
	num_attribs = ((unsigned int*)marshalled_TPH)[1];
	offset = 2*sizeof(unsigned int);

	for(i=0; i<num_stamps; i++) {
	    new_ps = COVEN_Profile_Stamp_Malloc();

	    new_ps->processor_number = ((unsigned int*)(marshalled_TPH+
	      offset))[0];
	    new_ps->mpirank = ((unsigned int*)(marshalled_TPH+offset))[1];
	    new_ps->coven_task_id = ((unsigned int*)(marshalled_TPH+offset))[2];
	    new_ps->TPH_num_tags = ((unsigned int*)(marshalled_TPH+offset))[3];
	    new_ps->TPH_num_total_bytes = ((unsigned int*)(marshalled_TPH+
	      offset))[4];
	    offset += 5*sizeof(unsigned int);
	    new_ps->stamp_type = ((_stamp_type*)(marshalled_TPH+offset))[0];
	    offset += sizeof(_stamp_type);
	    new_ps->special_int = ((unsigned int*)(marshalled_TPH+offset))[0];
	    offset += sizeof(unsigned int);
	    memcpy(new_ps->special_string, &((unsigned char*)(marshalled_TPH+
	      offset))[0], 128*sizeof(unsigned char));
	    offset += 128*sizeof(unsigned char);
	    new_ps->timestamp = ((struct timeval*)(marshalled_TPH+offset))[0];
	    offset += sizeof(struct timeval);

	    COVEN_Profile_Table_Append_Stamp(tph->pt, new_ps);
	}

	for(i=0; i<num_attribs; i++) {
	    taglen = ((unsigned int*)(marshalled_TPH+offset))[0];
	    atype = ((unsigned int*)(marshalled_TPH+offset))[1];
	    tag = (char*)COVEN_malloc(sizeof(char)*(taglen+1));
	    memcpy(tag, &((unsigned char*)(marshalled_TPH+offset))[2*
	      sizeof(unsigned int)], (taglen+1)*sizeof(unsigned char));
	    offset += 2*sizeof(unsigned int)+1+taglen;
	    switch(atype) {
		case COVEN_CHAR:
		    adata = (char*)COVEN_malloc(sizeof(char));
		    memcpy(adata, &((unsigned char*)(marshalled_TPH+
		      offset))[0], sizeof(char));
		    offset += sizeof(char);
		    break;
	        case COVEN_UCHAR: 
		    adata = (unsigned char*)COVEN_malloc(sizeof(unsigned char));
		    memcpy(adata, &((unsigned char*)(marshalled_TPH+offset))[0],
		      sizeof(unsigned char));
		    offset += sizeof(unsigned char);
		    break;
		case COVEN_SHORT: 
		    adata = (short*)COVEN_malloc(sizeof(short));
		    memcpy(adata, &((unsigned char*)(marshalled_TPH+offset))[0],
		      sizeof(short));
		    offset += sizeof(short);
		    break;
		case COVEN_USHORT:
		    adata = (unsigned short*)COVEN_malloc(sizeof(unsigned short));
		    memcpy(adata, &((unsigned char*)(marshalled_TPH+offset))[0],
		      sizeof(unsigned short));
		    offset += sizeof(unsigned short);
		    break;
		case COVEN_INT:
		    adata = (int*)COVEN_malloc(sizeof(int));
		    memcpy(adata, &((unsigned char*)(marshalled_TPH+offset))[0],
		      sizeof(int));
		    offset += sizeof(int);
		    break;
		case COVEN_UINT:
		    adata = (unsigned int*)COVEN_malloc(sizeof(unsigned int));
		    memcpy(adata, &((unsigned char*)(marshalled_TPH+offset))[0],
		      sizeof(unsigned int));
		    offset += sizeof(unsigned int);
		    break;
		case COVEN_FLOAT: 
		    adata = (float*)COVEN_malloc(sizeof(float));
		    memcpy(adata, &((unsigned char*)(marshalled_TPH+offset))[0],
		      sizeof(float));
		    offset += sizeof(float);
		    break;
		case COVEN_DOUBLE:
		    adata = (double*)COVEN_malloc(sizeof(double));
		    memcpy(adata, &((unsigned char*)(marshalled_TPH+offset))[0],
		      sizeof(double));
		    offset += sizeof(double);
		    break;
		case COVEN_STRING:
		    /* but for strings we have to know the length of the 
		     * string */
		    strlength = ((unsigned int*)(marshalled_TPH+offset))[0];
		    offset += sizeof(unsigned int);
		    adata = (unsigned char*)COVEN_malloc((strlength+1)*sizeof(
		      unsigned char));
		    memcpy(adata, &((unsigned char*)(marshalled_TPH+offset))[0],
		      (strlength+1)*sizeof(unsigned char));
		    offset += (strlength+1)*sizeof(unsigned char);
		    break;
		default: COVEN_Error("Unmarshall: Corrupt type");
	    }
	    new_at = COVEN_New_Attribute(atype, tag, adata);
	    /* the New_Attribute function WILL malloc the tag and data
	     * region, mainly because there are times when we want to 
	     * copy an attribute and want duplicates, not pointers to
	     * duplicates.  Therefore, we need to free these values here */
	    COVEN_debug(DMARSHALL," - unmarshall: new attr, freeing tag and data.\n");
	    COVEN_free(tag);
	    COVEN_free(adata);
	    // COVEN_Print_Attribute(new_at);
	    COVEN_Attribute_Table_Append_Attrib(tph->at, new_at);
	}
						
	numbufs = ((unsigned int*)(marshalled_TPH+offset))[0];

	offset += sizeof(unsigned int);

	for(i=0; i<numbufs; i++) {
	    int size;

	    taglen = ((unsigned int*)(marshalled_TPH+offset))[0];
	    ptype = ((unsigned int*)(marshalled_TPH+offset))[1];

	    if(ptype < PSE_TYPE_OFFSET) size = COVEN_get_type_size(ptype);
	    else size = PSE_get_type_size(ptype);

	    psize = ((unsigned int*)(marshalled_TPH+offset))[2];
	    seg_num = ((unsigned int*)(marshalled_TPH+offset))[3];
	    num_segs = ((unsigned int*)(marshalled_TPH+offset))[4];
	    tag = (char*)COVEN_malloc(sizeof(char)*(taglen+1));
	    memcpy(tag, &((unsigned char*)(marshalled_TPH+offset))[5*
	      sizeof(unsigned int)], taglen+1);
	    buf = (struct COVEN_Buffer*)COVEN_malloc(sizeof(struct COVEN_Buffer)+
	      psize*size);
	    buf->type = ptype;
	    buf->size = psize;
	    buf->seg_num = seg_num;
	    buf->num_segs = num_segs;
	    buf->tph = tph;
	    memcpy(COVEN_Buffer_Data(buf), 
	      &((unsigned char*)(marshalled_TPH+offset))[5*
	      sizeof(unsigned int)+taglen+1], psize * size);
	    buf->tag = (char*)COVEN_malloc((strlen(tag)+1)*sizeof(char));
	    strcpy(buf->tag,tag);
	    COVEN_Put_Buffer(tph, buf);
	    COVEN_free(tag);
	    offset += 5*sizeof(unsigned int) + 1 + taglen + psize * size;
	}

	COVEN_debug(DMARSHALL," - unmarshall TPH done\n");
	return tph;
}
