/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

#include "coven.h"
#include <stdlib.h>
#include <string.h>

int profiling_active = FALSE;

void COVEN_Profile_Enable(void) { profiling_active = TRUE; }
void COVEN_Profile_Disable(void) { profiling_active = FALSE; }

extern int myMPIrank, myCOVENTaskID, myProcessorNumber;

COVEN_Profile_Stamp * COVEN_Profile_Stamp_Malloc(void)
{  
	COVEN_Profile_Stamp *ps;
	if(profiling_active == FALSE) return NULL;
	ps = (COVEN_Profile_Stamp *)COVEN_malloc(sizeof(COVEN_Profile_Stamp));
	/* by default, we have no special string */
	ps->special_string[0] = '\0';
	return ps;
}

void COVEN_Profile_Stamp_Free(COVEN_Profile_Stamp *ps)
{
	if(profiling_active == FALSE) return;
	COVEN_free(ps);
}

COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Module(struct COVEN_TPH *tph,
	struct COVEN_Mod_Call *mod_call)
{
	COVEN_Profile_Stamp *ps;
	
	if(profiling_active == FALSE) return NULL;
	ps = COVEN_Profile_Stamp_Malloc();
	COVEN_Profile_Stamp_TPH(ps, tph);
	ps->special_int = mod_call->program_number;
	strncpy(ps->special_string, mod_call->module_name, 128);

	return ps;
}

COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Module_In(struct COVEN_TPH *tph,
	struct COVEN_Mod_Call *mod_call)
{
	COVEN_Profile_Stamp *ps;

	if(profiling_active == FALSE) return NULL;
	ps = COVEN_Profile_Stamp_TPH_Module(tph, mod_call);
	ps->stamp_type = Module_In;
	COVEN_Profile_Table_Append_Stamp(tph->pt, ps);

	return ps;
}

COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Module_Out(struct COVEN_TPH *tph,
	struct COVEN_Mod_Call *mod_call)
{
	COVEN_Profile_Stamp *ps;

	if(profiling_active == FALSE) return NULL;
	ps = COVEN_Profile_Stamp_TPH_Module(tph, mod_call);
	ps->stamp_type = Module_Out;
	COVEN_Profile_Table_Append_Stamp(tph->pt, ps);

	return ps;
}

COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Queue(struct COVEN_TPH *tph,
	struct COVEN_TPH_Queue *q)
{
	COVEN_Profile_Stamp *ps;
	
	if(profiling_active == FALSE) return NULL;
	ps = COVEN_Profile_Stamp_Malloc();
	COVEN_Profile_Stamp_TPH(ps, tph);
	ps->special_int = q->size;

	return ps;
}

COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Queue_In(struct COVEN_TPH *tph,
	struct COVEN_TPH_Queue *q)
{
	COVEN_Profile_Stamp *ps;

	if(profiling_active == FALSE) return NULL;
	ps = COVEN_Profile_Stamp_TPH_Queue(tph, q);
	ps->stamp_type = Queue_In;
	COVEN_Profile_Table_Append_Stamp(tph->pt, ps);

	return ps;
}

COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Queue_Out(struct COVEN_TPH *tph,
	struct COVEN_TPH_Queue *q)
{
	COVEN_Profile_Stamp *ps;

	if(profiling_active == FALSE) return NULL;
	ps = COVEN_Profile_Stamp_TPH_Queue(tph, q);
	ps->stamp_type = Queue_Out;
	/* we do this b/c right now we'd like to see the size of the 
	 * queue BEFORE we remove this TPH from the queue */
	ps->special_int++; 
	COVEN_Profile_Table_Append_Stamp(tph->pt, ps);

	return ps;
}

COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Created(struct COVEN_TPH *tph)
{
	COVEN_Profile_Stamp *ps;

	if(profiling_active == FALSE) return NULL;
	ps = COVEN_Profile_Stamp_Malloc();
	COVEN_Profile_Stamp_TPH(ps, tph);
	ps->stamp_type = Created;
	COVEN_Profile_Table_Append_Stamp(tph->pt, ps);

	return ps;
}

COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Destroyed(struct COVEN_TPH *tph)
{
	COVEN_Profile_Stamp *ps;

	if(profiling_active == FALSE) return NULL;
	ps = COVEN_Profile_Stamp_Malloc();
	COVEN_Profile_Stamp_TPH(ps, tph);
	ps->stamp_type = Destroyed;
	COVEN_Profile_Table_Append_Stamp(tph->pt, ps);

	return ps;
}

COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Sleep(struct COVEN_TPH *tph)
{
	COVEN_Profile_Stamp *ps;

	if(profiling_active == FALSE) return NULL;
	ps = COVEN_Profile_Stamp_Malloc();
	COVEN_Profile_Stamp_TPH(ps, tph);
	ps->stamp_type = Sleep;
	COVEN_Profile_Table_Append_Stamp(tph->pt, ps);

	return ps;
}

COVEN_Profile_Stamp * COVEN_Profile_Stamp_TPH_Awaken(struct COVEN_TPH *tph)
{
	COVEN_Profile_Stamp *ps;

	if(profiling_active == FALSE) return NULL;
	ps = COVEN_Profile_Stamp_Malloc();
	COVEN_Profile_Stamp_TPH(ps, tph);
	ps->stamp_type = Awaken;
	COVEN_Profile_Table_Append_Stamp(tph->pt, ps);

	return ps;
}

void COVEN_Profile_Stamp_TPH(COVEN_Profile_Stamp *ps, struct COVEN_TPH *tph)
{
	if(profiling_active == FALSE) return;
	ps->processor_number = myProcessorNumber;
	ps->mpirank = myMPIrank;
	ps->coven_task_id = myCOVENTaskID;
	gettimeofday(&(ps->timestamp), NULL);

	COVEN_TPH_Get_Stats(tph, &(ps->TPH_num_tags), 
	  &(ps->TPH_num_total_bytes));
}

void COVEN_Profile_Stamp_Copy(COVEN_Profile_Stamp *src, 
  COVEN_Profile_Stamp *dst)
{
	if(profiling_active == FALSE) return;
	dst->processor_number = src->processor_number;
	dst->mpirank = src->mpirank;
	dst->coven_task_id = src->coven_task_id;
	dst->TPH_num_tags = src->TPH_num_tags;
	dst->TPH_num_total_bytes = src->TPH_num_total_bytes;
	dst->stamp_type = src->stamp_type;
	dst->special_int = src->special_int;
	strcpy(dst->special_string, src->special_string);
	dst->timestamp = src->timestamp;
}

void COVEN_Profile_Table_Append_Stamp(COVEN_Profile_Table *pt, 
  COVEN_Profile_Stamp *new_ps)
{
	if(profiling_active == FALSE) return;
	new_ps->next = NULL;

	/* no head yet */
	if(pt->head == NULL) {
	    pt->head = new_ps;
  	    pt->tail = new_ps;
	    return;
	}

	/* else */
	pt->tail->next = new_ps;
	pt->tail = new_ps;
}

void COVEN_Profile_Table_Print(COVEN_Profile_Table *pt, FILE *stream)
{
	COVEN_Profile_Stamp *cur = pt->head;

	if(profiling_active == FALSE) return;
	printf(" -- COVEN_Profile\n");
	while(cur != NULL) {
	    fprintf(stream,
	      "%s  [%ld secs, %ld usecs] Proc# %d, MPITID# %d, COVENTID %d, "
	      "TPH#tags %d, TPH#bytes %d, SPint %d, SPstr `%s`\n",
	      COVEN_Profile_Stamp_Type_String(cur->stamp_type),
	      cur->timestamp.tv_sec,
	      cur->timestamp.tv_usec,
	      cur->processor_number,
	      cur->mpirank,
	      cur->coven_task_id,
	      cur->TPH_num_tags,
	      cur->TPH_num_total_bytes,
	      cur->special_int,
	      cur->special_string
	    );
	    cur = cur->next;
	}
	fflush(stream);
}

void COVEN_Profile_Table_Write(struct COVEN_TPH *tph)
{
	struct COVEN_Attribute *at;
	
	if(profiling_active == FALSE) return;
	at = COVEN_Attribute_Table_Find_Attrib(tph->at, 
	  "Working_Base_Directory");

	if(at != NULL) {
	    FILE *outfile;
	    char fn[128];
	    char * dir = (char *)COVEN_Attribute_Table_Get_Attrib_Data(tph->at,
	      "Working_Base_Directory");
	    printf("write to directory: %s\n", dir);
	    sprintf(fn, "%s/TPH-%x.prof", dir, (unsigned int)tph);
	    outfile = fopen(fn, "w");
	    COVEN_Profile_Table_Print(tph->pt, outfile);
	    fclose(outfile);
	}

	else {
	    fprintf(stderr,"Cannot write profile information, unsure of what "
	      "directory to write to.\n");
	}
}

void COVEN_Profile_Table_Free(COVEN_Profile_Table *pt)
{
	COVEN_Profile_Stamp *cur = pt->head;
	COVEN_Profile_Stamp *tmp;

	if(profiling_active == FALSE) return;
	while(cur != NULL) {
	    tmp = cur->next;
	    COVEN_Profile_Stamp_Free(cur);
	    cur = tmp;
	}
}

char * COVEN_Profile_Stamp_Type_String(_stamp_type st)
{
	if(profiling_active == FALSE) return NULL;
	switch(st) {
	    case Created:
		return "Created";
	    case Destroyed:
		return "Destroyed";
	    case Queue_In:
		return "Queue_In";
	    case Queue_Out:
		return "Queue_Out";
	    case Module_In:
		return "Module_In";
	    case Module_Out:
		return "Module_Out";
	    case Sleep:
		return "Sleep";
	    case Awaken:
		return "Awaken";
	}

	return "UNKNOWN_STAMP_TYPE";
}
