/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

#include "coven.h"
#include <stdlib.h>
#include <string.h>

extern int myMPIrank, myCOVENTaskID, myProcessorNumber;
extern int num_Master_Tasks, num_Slave_Tasks, num_Slave_Processors, am_slave;

extern struct COVEN_Mod_Calls *mod_calls;

int Scheduler_Next_Target = 1;

/* prints a TPH out for debugging purposes */
void COVEN_Print_TPH(struct COVEN_TPH *tph)
{
	struct COVEN_Buffer_List_El *bufel = tph->head;
	struct COVEN_Attribute *cur = tph->at->head;

	COVEN_debug(DTPH, "\t----{PRINTING TPH at address %x}----\n", 
	  (unsigned int)tph);
	while(cur) {
	    COVEN_debug(DTPH, "\t  ATTRIBUTE TAG: \n"); 
	    COVEN_Print_Attribute(cur);
	    cur = cur->next;
	}
	while(bufel) {
	    int size;

	    COVEN_debug(DTPH,"\t  BUFFER - at address: %x\n",
	      (unsigned int)((bufel->buffer)));
	    COVEN_debug(DTPH,"\t    TAG: (%x)%16s\n",
	      (unsigned int)((bufel->buffer->tag)), 
	      bufel->buffer->tag);

	    if(bufel->buffer->type < PSE_TYPE_OFFSET) 
	    	size = COVEN_get_type_size(bufel->buffer->type);
	    else size = PSE_get_type_size(bufel->buffer->type);
	
	    COVEN_debug(DTPH,"\t    NUMELEM: (%x)%5d\t",
	      (unsigned int)((bufel->buffer->size)), 
	      bufel->buffer->size);
	    COVEN_debug(DTPH,"SIZE: %5d\t",
	      bufel->buffer->size * size);
	    COVEN_debug(DTPH,"DATA@: %x\n",
	      (unsigned int)COVEN_Buffer_Data(bufel->buffer));
	    bufel = bufel->next;
	}
	COVEN_Profile_Table_Print(tph->pt, stdout);
	COVEN_debug(DTPH,"\t----{............}----\n");
}

void COVEN_Setup_TPH(struct COVEN_TPH *tph)
{
	memset(tph, 0, sizeof(struct COVEN_TPH));
	
	tph->pt = (COVEN_Profile_Table *)COVEN_malloc(sizeof(COVEN_Profile_Table));
	tph->pt->head = NULL;
	tph->pt->tail = NULL;

	tph->at = (struct COVEN_Attribute_Table*)COVEN_malloc(sizeof(
	  struct COVEN_Attribute_Table));
	tph->at->head = tph->at->tail = NULL;

	tph->head = NULL;
	tph->tail = NULL;
}

struct COVEN_TPH * COVEN_New_TPH(void)
{
	struct COVEN_TPH *tph;

	tph = (struct COVEN_TPH*)COVEN_malloc(sizeof(struct COVEN_TPH));
	COVEN_debug(DMEM,"  TPH malloc(): TPH @ address %x\n", (unsigned int)tph);
	COVEN_Setup_TPH(tph);
	return tph;
}

/* this is the basic copy TPH function, see COVEN_Copy_TPH_Exclude for
 * why we might want to exclude some buffers.  This excludes none (NULL) */
struct COVEN_TPH * COVEN_Copy_TPH(struct COVEN_TPH *tph)
{
	return COVEN_Copy_TPH_Exclude(tph, NULL, 0);
}

/* This performs the copy of one TPH into another, it copies the attribute
 * table (and attributes) as well as all the buffers.  Buffers can be
 * excluded by passing in an array of Collective_Actions (see header file).
 * These actions include buffers, in this case, if those buffers are in
 * this TPH, those are NOT coppied - excluded. */
struct COVEN_TPH * COVEN_Copy_TPH_Exclude(struct COVEN_TPH *tph,
  struct COVEN_Collective_Action *actions, int num_actions)
{
	struct COVEN_TPH *newTPH;
	struct COVEN_Buffer_List_El *bufel = tph->head;
	struct COVEN_Buffer *buf;
	struct COVEN_Attribute *cur = tph->at->head;
	struct COVEN_Attribute *new_at;
	int cur_action;
	/*
	COVEN_Profile_Stamp *cur_ps = tph->pt->head;
	COVEN_Profile_Stamp *new_ps;
	*/

	newTPH = COVEN_New_TPH();
	COVEN_Profile_Stamp_TPH_Created(newTPH);

	/*
	while(cur_ps) {
		new_ps = COVEN_Profile_Stamp_Malloc();
		COVEN_Profile_Stamp_Copy(cur_ps, new_ps);
		COVEN_Profile_Table_Append_Stamp(newTPH->pt, new_ps);
		cur_ps = cur_ps->next;
	}
	*/
	
	/* copy all the attributes */
	while(cur) {
	    //trace1("Copying attribute '%s'\n", cur->tag);
	    new_at = COVEN_New_Attribute(cur->attribute_type,
	      cur->tag, cur->data);
	    COVEN_Attribute_Table_Append_Attrib(newTPH->at, new_at);
	    cur = cur->next;
	}

	/* copy those buffers that are not excluded */
	while(bufel) {
	    int size;
	    int found = FALSE;

	    for(cur_action = 0; cur_action < num_actions && !found; 
	      cur_action++) 
	    {
	        /* exclude this guy */
	    	if(actions[cur_action].buffer == bufel->buffer) {
		    found = TRUE;
		}
	    }

	    if(!found) {
	    	if(bufel->buffer->type < PSE_TYPE_OFFSET) 
	    	    size = COVEN_get_type_size(bufel->buffer->type);
	    	else size = PSE_get_type_size(bufel->buffer->type);

	    	//trace2("Copying buffer %s to new TPH - size = %d bytes.\n", 
	    	//  bufel->tag,
	    	//  (bufel->buffer->size)*
		//    COVEN_TypeSizeArray[bufel->buffer->type]);
	    	buf = (struct COVEN_Buffer*)COVEN_malloc(sizeof(struct COVEN_Buffer)+
	      	  (bufel->buffer->size) * size);
	    	buf->type = bufel->buffer->type;
	    	buf->size = bufel->buffer->size;
	    	buf->tph = newTPH;
	    	memcpy(COVEN_Buffer_Data(buf), COVEN_Buffer_Data(bufel->buffer),
	    	  (bufel->buffer->size) * size);
	    	buf->tag = (char*)COVEN_malloc((strlen(bufel->buffer->tag)+1)*
	    	  sizeof(char));
	    	strcpy(buf->tag,bufel->buffer->tag);
	    	COVEN_Put_Buffer(newTPH, buf);
	    }
	    /* get the next buffer (element) */
 	    bufel = bufel->next;
	}

	return newTPH;
}

/* free's the buffers associated with a tph */
void COVEN_Free_TPH(struct COVEN_TPH *tph)
{
	struct COVEN_Buffer_List_El *bufel = tph->head;
	struct COVEN_Buffer_List_El *next;
	
	COVEN_debug(DMEM,"  TPH FREE(): TPH @ address %x\n", (unsigned int)tph);
	
	/* right now, let's print out the palantir numbers when the TPH is
	 * about to be dealloced */
	COVEN_Profile_Table_Write(tph);
	
	while(bufel) {
	    next = bufel->next;
	    /* here's the strange thing, a TPH may have a buffer element pointer
	     * to a buffer that it isn't considered the "owner" of, therefore
	     * we have to check this before we actually dealloc the 
	     * memory. :) */
	    if(bufel->buffer->tph == tph) {
		COVEN_debug(DTPH,"FREE(): BUF '%s' considered owned by %x "
		  "= %x, deallocing.\n",
		  (bufel->buffer->tag), (unsigned int)(bufel->buffer->tph), 
		  (unsigned int)tph);
		COVEN_Free_Buffer(bufel->buffer);
	    }
//	    else {
//			printf("=-=-=-=-=- BUFFER named '%s' NOT NOT NOT owned by %x = %x, NOT deallocing (someone else must own it).\n",
//				bufel->buffer->tag, bufel->buffer->tph, tph);
//		}
	    COVEN_debug(DTPH,"FREE(): freeing bufeln now.\n");
	    COVEN_debug(DTPH,"Free(): bufel at %x\n", (unsigned int)bufel);
	    COVEN_free(bufel); /* free the buffer element in the TPH */
	    COVEN_debug(DTPH,"FREE(): reloopin.\n");
	    bufel = next;
	}
	COVEN_debug(DTPH,"FREE(): Done w/ buffers, not prof tables.\n");
	COVEN_Profile_Table_Free(tph->pt);
	COVEN_Free_Attribute_Table(tph->at);
	COVEN_free(tph->pt);
	COVEN_free(tph->at);
	COVEN_free(tph);
}

/* this looks through the TPH and determines which data we can dealloc (if any)
 * now that we have completed stepnum */
void COVEN_Cleanup_TPH(struct COVEN_TPH *tph, int completed_stepnum, struct COVEN_Tag_Usage *tagusage)
{
	struct COVEN_Buffer_List_El *bufel = tph->head;
	struct COVEN_Buffer_List_El *next, *last;

	last = NULL;
	COVEN_debug(DTPH, "CLeanup: tph iS:\n");
	COVEN_Print_TPH(tph);
	while(bufel) {
	    COVEN_debug(DTPH, "CLeanup(): while....\n");
	    next = bufel->next;
	    /* we should only look at things that this TPH is considered 
	     * the owner of, see Free_TPH for explanation */
	    if(bufel->buffer->tph == tph) {
		int last_used = COVEN_Find_Last_Used(bufel->buffer->tag, 
		  tagusage);
		if(last_used != -1 && last_used <= completed_stepnum) {
		    trace2("STEP %d completed, we can now DEALLOC tag %s - "
		      "doing so now.\n",
		      completed_stepnum, bufel->buffer->tag);
		    if(last != NULL) last->next = bufel->next;
		    if(tph->head == bufel) tph->head = bufel->next;
		    COVEN_debug(DTPH, "FReeeeeeing BVUFER!\n");
		    COVEN_Free_Buffer(bufel->buffer);
		    COVEN_debug(DTPH, "FReed buffeR!!\n");
		    COVEN_free(bufel);
		}
		else last = bufel;
	    }
	    bufel = next;
	}
	COVEN_debug(DTPH, "Cleanup(): exitiN!\n");
}

/* writes an attribute which says that this TPH can be dealloced whenever
 * appropriate */
void COVEN_Schedule_TPH_For_Dealloc(struct COVEN_TPH *tph)
{
	struct COVEN_Attribute *at;
	int tmp = 1; /* true for dealloc */

	at = COVEN_New_Attribute(COVEN_INT, "TPH_Scheduled_For_Dealloc", &tmp);
	COVEN_Attribute_Table_Force_Attrib(tph->at, at);
}

/* returns 1 if the TPH is found to be scheduled for deallocation */
int COVEN_TPH_Scheduled_For_Dealloc(struct COVEN_TPH *tph)
{
	struct COVEN_Attribute *at;

	at = COVEN_Attribute_Table_Find_Attrib(tph->at, 
	  "TPH_Scheduled_For_Dealloc");
	if(at != NULL) {
	    int should_free = *(int*)COVEN_Attribute_Table_Get_Attrib_Data(
	      tph->at, "TPH_Scheduled_For_Dealloc");
	    if(should_free) return 1;
	}

	return 0;
}

struct COVEN_TPH * COVEN_Get_TPH(void *data) 
{
	return ((struct COVEN_Buffer *)((unsigned char*)data-(
	  sizeof(struct COVEN_Buffer)-sizeof(int))))->tph;
}

void COVEN_TPH_Clear_Manual_Relay(struct COVEN_TPH *tph)
{
	struct COVEN_Attribute *at;
	at = COVEN_Attribute_Table_Find_Attrib(tph->at, 
	  "_coven_manually_relayed");
	if(at != NULL)
	    COVEN_Attribute_Table_Delete_Attrib(tph->at, at);
}

int COVEN_TPH_Manually_Relayed(struct COVEN_TPH *tph)
{
	struct COVEN_Attribute *at;

	at = COVEN_Attribute_Table_Find_Attrib(tph->at, 
	  "_coven_manually_relayed");
	if(at != NULL) {
	    int relay = *(int*)COVEN_Attribute_Table_Get_Attrib_Data(tph->at,
	      "_coven_manually_relayed");
	    if(relay) return 1;
	}
	return 0;
}

void COVEN_Manually_Relay(struct COVEN_TPH *tph, int module_number)
{
	struct COVEN_Attribute *at;
	int proc_type, task_id, tmp;
	int shutdown = FALSE;

	COVEN_debug(DMAIN,"RELAY: trying to relay to module # %d\n",
	  module_number);
	/* we've got a potential problem in that the module number MAY be past
	 * the number of modules in the system.  This would be because we are
	 * looping to the end of the last module and the looping code assumes
	 * Relay(module_at_end_of_loop + 1) */
	/* Greater than or equal to because we do module numbers
	 * 0 -> num_mod_calls -1 */
	if(module_number >= mod_calls->num_mod_calls) {
	    task_id = mod_calls->mod_calls[mod_calls->num_mod_calls-1].
	      task_number;
	    shutdown = TRUE;
	}

	else task_id = mod_calls->mod_calls[module_number].task_number;
	//proc_type = am_slave;
	proc_type = mod_calls->mod_calls[module_number].slave_module;

	/* just using it as a temp variable signifying we man. relayed */
	tmp = 1; 
	at = COVEN_New_Attribute(COVEN_INT, "_coven_manually_relayed", &tmp);
	COVEN_Attribute_Table_Force_Attrib(tph->at, at);

	at = COVEN_New_Attribute(COVEN_INT, "_coven_target_module_number", 
	  &module_number);
	COVEN_Attribute_Table_Force_Attrib(tph->at, at);

	if(!shutdown) {
	    if(task_id == myCOVENTaskID && proc_type == am_slave) {
		COVEN_debug(DMAIN,"RELAY: tph is being relayed LOCALLY\n");
	    	COVEN_Relay_TPH_Locally(tph, proc_type, task_id);
	    }
 	    else {
	        COVEN_debug(DMAIN,"RELAY: relaying to proc %d task %d\n",
		  proc_type, task_id);
	    	COVEN_Relay_TPH(tph, proc_type, -1, task_id);
	    	/* this thing needs to be dealloced once we're done looking 
		 * at it */
		COVEN_debug(DMAIN,"RELAY: remotely relaying, scheduling for dealloc\n");
	    	COVEN_Schedule_TPH_For_Dealloc(tph);
	    }
	}
	else {
	    /* we need to shut down, we're trying to goto the module after the
	     * last module, so the program should end now */

	    COVEN_debug(DMAIN, "Relay Shutdown: because we tried to relay too "
	      "far.\n");
	    COVEN_Shut_Down();
	}
}

void COVEN_Relay_TPH_Simple(struct COVEN_TPH *tph)
{
	int proc_type, task_id;

	COVEN_Get_Next_Task(tph, &proc_type, &task_id);
	COVEN_Relay_TPH(tph, proc_type, -1, task_id);
	COVEN_Free_TPH(tph);
}

int COVEN_Scheduler_Get_Next(void)
{
	int tmp = Scheduler_Next_Target;
	Scheduler_Next_Target = (Scheduler_Next_Target % 
	  num_Slave_Processors) + 1;
	trace2("\t\tSCHEDULER: this time [%d] next time [%d]\n", tmp, 
	  Scheduler_Next_Target);
	return tmp;
}

int COVEN_Convert_To_MPI_TID(int proc_type, int proc_num, int task_id)
{
	int targetMPItaskID;

	/* if the sender doesn't know their task ID then we'll get it for 
	 * them */
	if(task_id == -1) task_id = myCOVENTaskID;

	targetMPItaskID = task_id - 1;
	if(proc_type == SLAVE) {
	    targetMPItaskID += ((proc_num - 1) * num_Slave_Tasks) + 
	      num_Master_Tasks;
	}

	return targetMPItaskID;
}

void COVEN_Relay_TPH_Locally(struct COVEN_TPH *tph, int proc_type, int task_id)
{
	int proc_num;

	int targetMPItaskID;
	
	/* if this is to be relayed to the master, then we KNOW the processor 
	 * number it must be processor #0 - so set it! */
	if(proc_type == MASTER) proc_num = 0;
	else proc_num = myProcessorNumber;
	
	/* we need to reverse-engineer how to convert a processor # and a COVEN 
	 * Task ID into the original MPI task ID - do that here */
	targetMPItaskID = COVEN_Convert_To_MPI_TID(proc_type, proc_num, 
	  task_id);

	trace5("[COVEN_RELAY_TPH_locally] MPI%d -> proctype %d, proc_num %d, "
	  "taskid %d --> MPItask #%d\n",
	  myMPIrank, proc_type, proc_num, task_id, targetMPItaskID);

	{
	    /* enqueue this TPH on a local queue and send a signal letting 
	     * myself now that there are TPHs in a local queue to deal with */
	    int message = COVEN_TPH_QUEUED_MESSAGE;
	    COVEN_Enqueue_TPH(tph);
	    MPI_Send(&message, 1, MPI_BYTE, targetMPItaskID, 
	      COVEN_PROGRAM_SEQUENCER_TAG, MPI_COMM_WORLD);
	}
}

void COVEN_Relay_TPH(struct COVEN_TPH *tph, int proc_type, int proc_num, 
  int task_id)
{
	int targetMPItaskID;
	
	/* if there is no specified target processor number then we try and 
	 * come up with one */
	if(proc_num == -1) {
	    /* if this is to be relayed to the master, then we KNOW the 
	     * processor number it must be processor #0 - so set it! */
	    if(proc_type == MASTER) proc_num = 0;

	    /* else, we need to come up with a way to schedule which proc 
	     * it goes to */
	    else if(myProcessorNumber == 0) {
		proc_num = COVEN_Scheduler_Get_Next();
	    }
	    else {
		proc_num = myProcessorNumber;
	    }
	}
	
	/* we need to reverse-engineer how to convert a processor # and a 
	 * COVEN Task ID into the original MPI task ID - do that here */
	targetMPItaskID = COVEN_Convert_To_MPI_TID(proc_type, proc_num, 
	  task_id);

	COVEN_debug(DMAIN, "[COVEN_RELAY_TPH] MPI%d -> proctype %d, proc_num "
	  "%d, taskid %d --> MPItask #%d\n",
	  myMPIrank, proc_type, proc_num, task_id, targetMPItaskID);

	{
	    void *marshalled_TPH;
	    int marshalled_TPH_size;

	    COVEN_Marshall_TPH(tph, &marshalled_TPH, &marshalled_TPH_size);

	    trace1("\t->starting to send %d bytes!!\n", marshalled_TPH_size);

	    /* I -THINK- I can do a blocking Send here b/c it will return 
	     * after copying the data to a local buffer */
	    MPI_Send(marshalled_TPH, marshalled_TPH_size, MPI_BYTE,
	      targetMPItaskID, COVEN_PROGRAM_SEQUENCER_TAG,
	      MPI_COMM_WORLD);

	    trace("\t->send completed!!!!\n");

	    COVEN_free(marshalled_TPH);
	}
}

void COVEN_TPH_Get_Stats(struct COVEN_TPH *tph, int *num_tags, int *total_bytes)
{
	struct COVEN_Buffer_List_El *bufel = tph->head;

	*num_tags = 0;
	*total_bytes = 0;

	while(bufel) {
	    int size;

	    if(bufel->buffer->type < PSE_TYPE_OFFSET) 
	    	size = COVEN_get_type_size(bufel->buffer->type);
	    else size = PSE_get_type_size(bufel->buffer->type);

	    *num_tags = *num_tags + 1;
	    *total_bytes += bufel->buffer->size * size;
	    bufel = bufel->next;
	}
}
