/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

#include <math.h>
#include <stdlib.h>
#include "coven.h"
#include "medea.h"

/* this is stuff that's declared in coven_main.c and other compiled in
 * objects use externs to them, they're not important but need to be
 * linked in */
int myMPIrank;
int myCOVENTaskID = -1;
int myProcessorNumber;
int am_slave;
int num_Slave_Processors;
MPI_Comm MPI_COMM_THIS_PARALLEL_TASK;
struct COVEN_Tag_Usage *tagusage = NULL;
/* end */

struct COVEN_Mod_Binding binding[] = {
	{"my_bodies", "__NBODY_GENERATE_PLANAR_RANDY_1_output1"},
};

struct COVEN_Mod_Binding_Table btable_real = {
	1, 0, binding
};

int main(int argc, char **argv)
{
	struct COVEN_TPH *_default_tph;
	struct COVEN_Mod_Binding_Table *btable = &btable_real;

	int num_slave_procs = 2;

	_default_tph = COVEN_New_TPH();
	//COVEN_New_Buffer(_default_tph, 
	//  "__NBODY_GENERATE_PLANAR_RANDY_1_output1", 1000, MEDEA_BODY);
	COVEN_Print_TPH(_default_tph);

	{
//	    MEDEA_BODY_TYPE abodies = COVEN_Get_Buffer_Arg(_default_tph, 
//	      btable, "abodies", COVEN_OUTPUT_MODE, MEDEA_BODY, 666);
	    MEDEA_BODY_TYPE my_bodies = COVEN_Get_Buffer_Arg(_default_tph, 
	      btable, "my_bodies", COVEN_OUTPUT_MODE, MEDEA_BODY, 9);
//	    MEDEA_BODY_TYPE bbodies = COVEN_Get_Buffer_Arg(_default_tph, 
//	      btable, "bbodies", COVEN_OUTPUT_MODE, MEDEA_BODY, 98765);
  	    int total_num_bodies = COVEN_COUNT(my_bodies);

	    int target_num_bodies = (int)ceil((double)total_num_bodies / 
	      (double)num_slave_procs);

	    //struct COVEN_TPH *newTPHs;
	    int i, total_assigned_bodies;
	    COVEN_size *array_of_blocks, *array_of_displacements;
	    COVEN_Type body_type, scatter_type;

	    COVEN_Print_TPH(_default_tph);
	    COVEN_Type_contiguous(sizeof(struct Medea_Body), COVEN_BYTE, 
	      &body_type);

	    printf("Total num bodies = %d\n", total_num_bodies);
	    printf("Target num bodies = %d\n", target_num_bodies);
	    printf("Address of my_bodies = %x\n", my_bodies);
	    total_assigned_bodies = 0;
	    i = 0;

	    array_of_blocks = (COVEN_size*)malloc(num_slave_procs*
	      sizeof(COVEN_size));
	    array_of_displacements = (COVEN_size*)malloc(num_slave_procs*
	      sizeof(COVEN_size));
	
	    while(total_assigned_bodies < total_num_bodies) {
	    	/* try and fit this many bodies */
	    	int num_bodies = target_num_bodies;
	    	/* starting at how many we've assigned so far */
	    	int start_body = total_assigned_bodies;

	    	total_assigned_bodies += num_bodies;
	    	/* if we've assigned more than we have, then back off by that
	    	 * amount */
	    	if(total_assigned_bodies > total_num_bodies)
	    	    num_bodies -= (total_assigned_bodies - total_num_bodies);

	    	/* now num_bodies is the number of bodies for this TPH to 
		 * take */
	    	array_of_blocks[i] = num_bodies;
	    	array_of_displacements[i] = start_body;
	    	i++;
	    }

	    for(i=0; i<num_slave_procs; i++) {
	    	printf("blocks[%d] = %lld, displacements[%d] = %lld\n",
		  i, array_of_blocks[i], i, array_of_displacements[i]);
	    }

	    /* now that we have our displacements and blocks setup we can make
	     * the MPI data type */
	    COVEN_Type_indexed(num_slave_procs, array_of_blocks, 
	      array_of_displacements, body_type, &scatter_type);
	    printf("type scatter created\n");

	    {
	    	struct COVEN_Collective_Action *col_action;

	    	col_action = (struct COVEN_Collective_Action*)malloc(1*
	      	  sizeof(struct COVEN_Collective_Action));
	    	COVEN_Create_Scatter_Action(&(col_action[0]), scatter_type, 
	      	  "my_bodies");
		printf("\n**** BEGIN: COVEN_Scatter\n\n");
	    	COVEN_Scatter(num_slave_procs, col_action, 1);
		printf("\n**** END: COVEN_Scatter\n\n");
	    }

	    free(array_of_blocks);
	    free(array_of_displacements);
	}
	return 0;
}
